/*
* OpenGL - 3D Vision Wrapper
* Copyright (c) Helifax 2015
*/

#ifndef __CONFIG_READER_H__
#define __CONFIG_READER_H__

#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <vector>
#include "windows.h"
#include "ShaderManager.h"

//Reads the basic information
// NOT the shader exceptions

typedef struct  
{
	int altKeyCode;
	float altConvergence;
	float altSeparation;
	bool enableToggleMode;
	CUSTOM_SHADER_VALUES_T customShaderValues;
	bool customShaderisApplied;
	int keyState;
	int keyPrevState;
} ALT_KEYS_T;

// Selects how the L/R eyes are shifted/generated.
typedef enum 
{
	DOMINANT_LEFT_EYE,
	DOMINANT_RIGHT_EYE,
	DOMINANT_MONO_EYE,
} DOMINANT_EYE_T;

class configReader
{
public:
	//ctor
	configReader();
	~configReader();
	// ---------------------------------------------------------------------------------------------

	bool GetDevelopmentMode()
	{
		return m_devMode;
	}
	// ---------------------------------------------------------------------------------------------

	 bool GetEnableLog()
	{
		return m_enableLog;
	}
	// ---------------------------------------------------------------------------------------------
	 bool GetEnableSplashScreen()
	{
		return m_enableSplashScreen;
	}
	// ---------------------------------------------------------------------------------------------

	 bool GetEnableShaderDump()
	{
		return m_enableShaderDump;
	}
	// ---------------------------------------------------------------------------------------------

	 bool GetVertexStereoInjection()
	{
		return m_enableVertexStereoInjection;
	}
	// ---------------------------------------------------------------------------------------------

	 bool isLegacyOpenGLEnabled()
	 {
		 return m_legacyOpenGL;
	 }
	 // ---------------------------------------------------------------------------------------------

	 float GetLegacyHUDSeparation()
	 {
		 return m_legacyHUDSeparation;
	 }
	 // ---------------------------------------------------------------------------------------------

	 bool GetKeyToggleMode(unsigned int index)
	{
		if (index < m_altKeys.size())
			return m_altKeys[index].enableToggleMode;
		else
			return FALSE;
	}
	// ---------------------------------------------------------------------------------------------

	 unsigned int GetNumberOfKeys()
	 {
		 return m_altKeys.size();
	 }
	 // ---------------------------------------------------------------------------------------------

	 int GetKeyNumber(unsigned int index)
	{
		if (index < m_altKeys.size())
			return m_altKeys[index].altKeyCode;
		else
			return -1;
	}
	// ---------------------------------------------------------------------------------------------

	 float GetAltConvergence(unsigned int index)
	 {
		 if (index < m_altKeys.size())
			 return m_altKeys[index].altConvergence;
		 else
			 return -1;
	}
	// ---------------------------------------------------------------------------------------------

	 float GetAltSeparation(unsigned int index)
	 {
		 if (index < m_altKeys.size())
			 return  m_altKeys[index].altSeparation;
		 else
			 return -1;
	 }
	 // ---------------------------------------------------------------------------------------------

	 BOOL GetAltCustomShaderParams(unsigned int index, CUSTOM_SHADER_VALUES_T &customValues);
	 bool GetAltCustomShaderIsAnythingApplied(void);
	 bool GetAltCustomShaderApplied(unsigned int index);
	 void SetAltCustomShaderApplied(unsigned int index, bool state);
	 // ---------------------------------------------------------------------------------------------

	 int GetKeyState(unsigned int index)
	 {
		 if (index < m_altKeys.size())
			 return  m_altKeys[index].keyState;
		 else
			 return 0;
	 }
	 // ---------------------------------------------------------------------------------------------

	 void SetKeyState(unsigned int index, int state)
	 {
		 if (index < m_altKeys.size())
			 m_altKeys[index].keyState = state;
	 }
	 // ---------------------------------------------------------------------------------------------

	 int GetKeyPrevState(unsigned int index)
	 {
		 if (index < m_altKeys.size())
			 return  m_altKeys[index].keyPrevState;
		 else
			 return 0;
	 }
	 // ---------------------------------------------------------------------------------------------

	 void SetKeyPrevState(unsigned int index, int state)
	 {
		 if (index < m_altKeys.size())
			 m_altKeys[index].keyPrevState = state;
	 }
	 // ---------------------------------------------------------------------------------------------

	float GetDepthFactor()
	{
		return m_depthFactor;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool Get3DVisionEnabledStatus()
	{
		return m_3dVisionEnabled;
	}
	// ---------------------------------------------------------------------------------------------

	inline void Set3DVisionEnabledStatus(bool newStatus)
	{
		m_3dVisionEnabled = newStatus;
	}
	// ---------------------------------------------------------------------------------------------

	bool Get3DCompatModeEnabled();
	// ---------------------------------------------------------------------------------------------

	bool Get3DCompatModeActive();
	// ---------------------------------------------------------------------------------------------

	inline void Set3DCompatModeActive(bool state)
	{
		m_is3DCompatModeActive = state;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool GetProxyLibGeneralistic()
	{
		return m_isProxyLibGeneralistic;
	}
	// ---------------------------------------------------------------------------------------------

	inline DOMINANT_EYE_T GetStereoDominantEye()
	{
		return m_dominantEye;
	}
	// ---------------------------------------------------------------------------------------------

	inline std::string Get3DCompatModeProxyLib()
	{
		return m_3DReshadeProxyLib;
	}
	// ---------------------------------------------------------------------------------------------
	
	inline bool GetSyncToRefreshRate()
	{
		return m_syncToScreenRefreshRate;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool GetAlternateFrames()
	{
		return m_alternateFrames;
	}
	// ---------------------------------------------------------------------------------------------
	
	inline void SetAlternateFrames(bool value)
	{
		m_alternateFrames = value;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool IsFullScreenForced()
	{
		return m_forceFullScreen;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool IsWindowModeSupportEnabled()
	{
		return m_windowModeEnabled;
	}
	// ---------------------------------------------------------------------------------------------

	inline bool IsFullScreenDetectionEnabled()
	{
		return m_fullscreenDetection;
	}
	// ---------------------------------------------------------------------------------------------

	unsigned int GetDefaultConvergence()
	{
		return m_defaultConvergence;
	}
	// ---------------------------------------------------------------------------------------------

	bool IsNVProfileForced()
	{
		return m_forceNvProfileLoad;
	}
	// ---------------------------------------------------------------------------------------------

	float Get3DVisionRating()
	{
		return m_3DVisionRating;
	}
	// ---------------------------------------------------------------------------------------------

private:
	//General
	bool m_enableSplashScreen;
	bool m_enableLog;
	bool m_enableShaderDump;
	bool m_devMode;
	bool m_is3DCompatModeEnabled;
	bool m_is3DCompatModeActive;
	bool m_isProxyLibGeneralistic;
	std::string m_3DReshadeProxyLib;
	DOMINANT_EYE_T m_dominantEye;
	
	// Insertion Point		
	bool m_3dVisionEnabled;
	bool m_syncToScreenRefreshRate;
	bool m_alternateFrames;
	bool m_forceFullScreen;
	bool m_windowModeEnabled;
	bool m_fullscreenDetection;	
	bool m_enableVertexStereoInjection;

	// OpenGL Legacy Code Flag
	bool m_legacyOpenGL;
	float m_legacyHUDSeparation;

	//3D Settings specific
	float m_depthFactor;
	unsigned int m_defaultConvergence;
	bool m_forceNvProfileLoad;
	float m_3DVisionRating;
	std::vector<ALT_KEYS_T> m_altKeys;

};


#endif