#include <sstream>
#include "ConfigReader.h"

extern HMODULE hProxyLib;

static std::string GetPath() 
{
	char buffer[MAX_PATH];
	GetModuleFileName(NULL, buffer, MAX_PATH);
	std::string::size_type pos = std::string(buffer).find_last_of("\\/");
	return std::string(buffer).substr(0, pos);	
}
// -----------------------------------------------------------------------------------

configReader::configReader()
{
	m_enableLog = false;
	m_enableShaderDump = false;
	m_enableVertexStereoInjection = false;
	m_is3DCompatModeActive = false;
	std::string configFileName = GetPath() + "/3DVisionWrapper.ini";
	std::ifstream configFile(configFileName);
	std::string configLine;

	//file is opened
	if (configFile.is_open())
	{
		// read one line at a time
		while (getline(configFile, configLine))
		{
			//----------------------------------------------
			// Find the log information
			if (configLine.find("[General]") != std::string::npos)
			{
				//get the next line
				while (configLine.find("EnableSplashScreen") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_enableSplashScreen = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_enableSplashScreen = false;
					}
				}

				//get the next line
				while (configLine.find("EnableLog") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_enableLog = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_enableLog = false;
					}
				}

				while (configLine.find("EnableShaderDump") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_enableShaderDump = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_enableShaderDump = false;
					}
				}

				while (configLine.find("EnableDevMode") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_devMode = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_devMode = false;
					}
				}
			}
			//----------------------------------------------

			// [InsertionPoint]
			size_t foundLocation;
			if (configLine.find("[3D_Vision_Global_Settings]") != std::string::npos)
			{
				while ((foundLocation = configLine.find("Enable3DVision")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_3dVisionEnabled = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_3dVisionEnabled = false;
					}
				}
				while ((foundLocation = configLine.find("SyncToMonitorRefreshRate")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_syncToScreenRefreshRate = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_syncToScreenRefreshRate = false;
					}
				}
				while ((foundLocation = configLine.find("ForceFullScreen")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_forceFullScreen = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_forceFullScreen = FALSE;
					}
				}
				while ((foundLocation = configLine.find("ForceWindowModeSupport")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_windowModeEnabled = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_windowModeEnabled = FALSE;
					}
				}
				while ((foundLocation = configLine.find("FullScreenDetection")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_fullscreenDetection = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_fullscreenDetection = FALSE;
					}
				}
				while ((foundLocation = configLine.find("ForceNVProfileLoad")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_forceNvProfileLoad = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_forceNvProfileLoad = FALSE;
					}
				}
				while ((foundLocation = configLine.find("3DVisionRating")) == std::string::npos)
				{
					getline(configFile, configLine);
					size_t position = configLine.find("= ");
					std::string value = configLine.substr(position + 1);

					// Read the value
					m_3DVisionRating = std::stof(value, 0);
				}
				// Compatibility Mode 3D
				while ((foundLocation = configLine.find("EnableReshadeDepth3D")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_is3DCompatModeEnabled = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_is3DCompatModeEnabled = FALSE;
					}
				}
				// Compatibility Mode 3D Enabled by default
				while ((foundLocation = configLine.find("ReshadeDepth3D_Default_On")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_is3DCompatModeActive = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_is3DCompatModeActive = FALSE;
					}
				}

				// What type of library do we use?
				while ((foundLocation = configLine.find("ProxyLibGeneralistic")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_isProxyLibGeneralistic = TRUE;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_isProxyLibGeneralistic = FALSE;
					}
				}

				// Compatibility Mode 3D Library
				while ((foundLocation = configLine.find("ReshadeProxyLib")) == std::string::npos)
				{
					getline(configFile, configLine);
					// store the injection point
					size_t startPoint = configLine.find("\"", 0);
					size_t endPoint = configLine.find("\"", startPoint + 1);
					m_3DReshadeProxyLib = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
				}

				while ((foundLocation = configLine.find("StereoDominantEye")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("LEFT_EYE") != std::string::npos)
					{
						m_dominantEye = DOMINANT_LEFT_EYE;
					}
					else if (configLine.find("RIGHT_EYE") != std::string::npos)
					{
						m_dominantEye = DOMINANT_RIGHT_EYE;
					}
					else if (configLine.find("MONO_EYE") != std::string::npos)
					{
						m_dominantEye = DOMINANT_MONO_EYE;
					}
				}

				while ((foundLocation = configLine.find("AlternateFramesRendering")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_alternateFrames = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_alternateFrames = false;
					}
				}
				while ((foundLocation = configLine.find("DepthMultiplicationFactor")) == std::string::npos)
				{
					getline(configFile, configLine);
					size_t position = configLine.find("= ");
					std::string value = configLine.substr(position + 1);

					// Read the depth hack
					m_depthFactor = std::stof(value, 0);
					
					// If we use the dominant eye approach because we will only add stereo to one of the eyes we will loose 50% of depth.
					// In order to fix this we need to apply double the depth to the other eye.
					if (m_dominantEye != DOMINANT_MONO_EYE)
					{
						m_depthFactor *= 2;
					}
				}
				while ((foundLocation = configLine.find("DefaultConvergence")) == std::string::npos)
				{
					getline(configFile, configLine);
					size_t position = configLine.find("= ");
					std::string value = configLine.substr(position + 1);

					// Read the default convergence
					std::stringstream ss;
					ss << std::hex << value;
					ss >> m_defaultConvergence;
				}
			}

			// [Legacy_OpenGL_Calls]
			if (configLine.find("[Legacy_OpenGL_Calls]") != std::string::npos)
			{
				while ((foundLocation = configLine.find("LegacyMode")) == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("false") != std::string::npos)
					{
						m_legacyOpenGL = false;
					}
					else if (configLine.find("true") != std::string::npos)
					{
						m_legacyOpenGL = true;
					}
				}

				while ((foundLocation = configLine.find("LegacyHUDSeparation")) == std::string::npos)
				{
					getline(configFile, configLine);
					size_t position = configLine.find("= ");
					std::string value = configLine.substr(position + 1);

					// Read the value
					m_legacyHUDSeparation = std::stof(value, 0);
				}
			}

			//[Alternative_3D_Settings]
			if (configLine.find("[Alternative_3D_Settings]") != std::string::npos)
			{
				while (configLine.find("[End]") == std::string::npos)
				{
					getline(configFile, configLine);
					//----------------------------------------------
					//get the next line
					if (configLine.find("NewKeyShortcut") != std::string::npos)
					{
						ALT_KEYS_T newKey;

						// store the injection point
						size_t startPoint = configLine.find("(", 0);
						size_t middlePoint1 = configLine.find(",", startPoint + 1);
						size_t middlePoint2 = configLine.find(",", middlePoint1 + 1);
						size_t middlePoint3 = configLine.find(",", middlePoint2 + 1);
						size_t middlePoint4 = configLine.find(",", middlePoint3 + 1);
						size_t middlePoint5 = configLine.find(",", middlePoint4 + 1);
						size_t middlePoint6 = configLine.find(",", middlePoint5 + 1);
						size_t middlePoint7 = configLine.find(",", middlePoint6 + 1);
						size_t endPoint = configLine.find(")", middlePoint3 + 1);

						// Read the key
						std::string temp = configLine.substr(startPoint + 1, middlePoint1 - startPoint - 1).c_str();
						newKey.altKeyCode = std::stoul(temp, nullptr, 16);
						newKey.altSeparation = (float)atof(configLine.substr(middlePoint1 + 1, middlePoint2 - middlePoint1 - 1).c_str());
						newKey.altConvergence = (float)atof(configLine.substr(middlePoint2 + 1, middlePoint3 - middlePoint2 - 1).c_str());
						newKey.customShaderValues.x = (float)atof(configLine.substr(middlePoint3 + 1, middlePoint4 - middlePoint3 - 1).c_str());
						newKey.customShaderValues.y = (float)atof(configLine.substr(middlePoint4 + 1, middlePoint5 - middlePoint4 - 1).c_str());
						newKey.customShaderValues.z = (float)atof(configLine.substr(middlePoint5 + 1, middlePoint6 - middlePoint5 - 1).c_str());
						newKey.customShaderValues.w = (float)atof(configLine.substr(middlePoint6 + 1, middlePoint7 - middlePoint6 - 1).c_str());
						temp = (configLine.substr(middlePoint7 + 1, endPoint - middlePoint7 - 1).c_str());

						if (temp.find("false") != std::string::npos)
							newKey.enableToggleMode = false;
						else if (configLine.find("true") != std::string::npos)
							newKey.enableToggleMode = true;

						// Set the initial states
						newKey.customShaderisApplied = false;
						newKey.keyPrevState = false;
						newKey.keyState = false;

						// Push them
						m_altKeys.push_back(newKey);
						//----------------------------------------------
					}
				}
			}
		}
		configFile.close();
	}
}
///-------------------------------------------------------------------------------------------

BOOL configReader::GetAltCustomShaderParams(unsigned int index, CUSTOM_SHADER_VALUES_T &customValues)
{
	if (index < m_altKeys.size())
	{
		customValues = m_altKeys[index].customShaderValues;
		return TRUE;
	}
	else
		return FALSE;
}
// ---------------------------------------------------------------------------------------------

bool configReader::GetAltCustomShaderIsAnythingApplied(void)
{
	for (size_t index = 0; index < m_altKeys.size(); index++)
	{
		if (m_altKeys[index].customShaderisApplied == true)
			return true;
	}
	return false;
}
// ---------------------------------------------------------------------------------------------

bool configReader::GetAltCustomShaderApplied(unsigned int index)
{
	if (index < m_altKeys.size())
	{
		return m_altKeys[index].customShaderisApplied;
	}
	return false;
}
// ---------------------------------------------------------------------------------------------

void configReader::SetAltCustomShaderApplied(unsigned int index, bool state)
{
	if (index < m_altKeys.size())
	{
		m_altKeys[index].customShaderisApplied = state;
	}
}
// ---------------------------------------------------------------------------------------------

bool configReader::Get3DCompatModeEnabled()
{
	return (m_is3DCompatModeEnabled && hProxyLib);
}
// ---------------------------------------------------------------------------------------------

bool configReader::Get3DCompatModeActive()
{
	return (m_is3DCompatModeActive && hProxyLib);
}
// ---------------------------------------------------------------------------------------------