#include "include\ShaderManager.h"
#include "include\configReader.h"
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <direct.h>
#include "opengl32.h"
#include "crc32.h"
#include <mutex>

extern "C" {
#include "opengl_3dv.h"
}


extern funct_glGetShaderiv_t orig_glGetShader;
extern funct_glGetShaderiv_t orig_glGetShader;
extern funct_glShaderSource_t orig_glShaderSource;
static bool m_allShadersApplied = false;
static bool m_allShaderPairsApplied = false;
extern configReader *g_reader;

extern int debugVertexIndex;
extern int debugPixelIndex;

static std::string GetPath()
{
	char buffer[MAX_PATH];
	GetModuleFileName(NULL, buffer, MAX_PATH);
	std::string::size_type pos = std::string(buffer).find_last_of("\\/");
	return std::string(buffer).substr(0, pos);
}
// -----------------------------------------------------------------------------------

static std::string removeDoubleSlash(std::string _input)
{
	size_t location =_input.find("\\n");
	while (location != std::string::npos)
	{
		_input.replace(location, 2, "\n");
		location = _input.find("\\n");
	}
	return _input;
}
//--------------------------------------------------------------------------------------------

ShaderManager *ShaderManager::instance = 0;

ShaderManager::ShaderManager()
{
	LoadCustomShaders();
	ReadConfigFile();
}
///-------------------------------------------------------------------------------------------

void ShaderManager::ReadConfigFile(void)
{
	// Remove the current values so we use the new ones
	m_customShaderValues.resize(0);

	// Read our shader information
	std::string configFileName = GetPath() + "/3DVisionWrapper.ini";
	std::ifstream configFile(configFileName);
	std::string configLine;

	// Default inits
	m_enableStereoVertexInjection = false;
	m_enableStereoFragmentInjection = false;
	m_vertexAutoStereoInjectionInfo = false;
	m_vertexAutoStereoInjectionInfo = false;

	//file is opened
	if (configFile.is_open())
	{
		// read one line at a time
		while (getline(configFile, configLine))
		{
			//----------------------------------------------
			// Find the Vertex Stereo Injection Point
			if (configLine.find("[Shader_Stereo_Injection]") != std::string::npos)
			{
				size_t startPoint;
				size_t endPoint;
				//----------------------------------------------
				//get the next line
				while (configLine.find("Enable") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_enableStereoVertexInjection = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_enableStereoVertexInjection = false;
					}
				}
				//----------------------------------------------
				//get the next line
				while (configLine.find("EnableUnity5") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_enableUnity5Mode = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_enableUnity5Mode = false;
					}
				}
				//----------------------------------------------
				//get the next line
				while (configLine.find("EmbedProgramIdInCRC32") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_embedProgInCRC = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_embedProgInCRC = false;
					}
				}
				//----------------------------------------------
				//get the next line
				while (configLine.find("VertexAutoStereoInjectionInfo") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_vertexAutoStereoInjectionInfo = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_vertexAutoStereoInjectionInfo = false;
					}
				}
				//----------------------------------------------
				//get the next line
				for (size_t i = 0; i < MAX_UNIFORM_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "UniformInjection", i+1);
					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						// store the injection point
						startPoint = configLine.find("\"", 0);
						endPoint = configLine.find("\"", startPoint + 1);
						m_UniformPoint[i] = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
					}
				}

				//----------------------------------------------
				//get the next line
				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "VertexInjectionPoint", i + 1);

					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						// store the injection point
						startPoint = configLine.find("\"", 0);
						endPoint = configLine.find("\"", startPoint + 1);
						m_vertexInjectionPoint[i] = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
					}
				}

				//----------------------------------------------
				//get the next line
				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "VertexStereoString", i + 1);

					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						// store the injection point
						startPoint = configLine.find("\"", 0);
						endPoint = configLine.find("\"", startPoint + 1);
						m_vertexStereoInjectionString[i] = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
						m_vertexStereoInjectionString[i] = removeDoubleSlash(m_vertexStereoInjectionString[i]);
					}
				}

				//get the next line
				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "VertexOperation", i + 1);

					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						if (configLine.find("insert") != std::string::npos)
						{
							m_vertexIsInsert[i] = true;
						}
						else if (configLine.find("replace") != std::string::npos)
						{
							m_vertexIsInsert[i] = false;
						}
					}
				}

				//----------------------------------------------
				//get the next line
				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "PixelInjectionPoint", i + 1);

					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						// store the injection point
						startPoint = configLine.find("\"", 0);
						endPoint = configLine.find("\"", startPoint + 1);
						m_pixelInjectionPoint[i] = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
					}
				}
				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "PixelStereoString", i + 1);

					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						// store the injection point
						startPoint = configLine.find("\"", 0);
						endPoint = configLine.find("\"", startPoint + 1);
						m_pixelStereoInjectionString[i] = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
						m_pixelStereoInjectionString[i] = removeDoubleSlash(m_pixelStereoInjectionString[i]);
					}
				}

				for (size_t i = 0; i < MAX_VERTEX_INJECTION_POINTS; i++)
				{
					char currentInjector[255];
					sprintf_s(currentInjector, "%s%d", "PixelOperation", i + 1);

					//get the next line
					while (configLine.find(currentInjector) == std::string::npos)
					{
						getline(configFile, configLine);
						if (configLine.find("insert") != std::string::npos)
						{
							m_pixelIsInsert[i] = true;
						}
						else if (configLine.find("replace") != std::string::npos)
						{
							m_pixelIsInsert[i] = false;
						}
					}
				}

				//get the next line
				while (configLine.find("ShaderSwapIsContinous") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_swapIsContinous = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_swapIsContinous = false;
					}
				}

				//get the next line
				while (configLine.find("ShaderPairSwapIsContinous") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("true") != std::string::npos)
					{
						m_swapPairIsContinous = true;
					}
					else if (configLine.find("false") != std::string::npos)
					{
						m_swapPairIsContinous = false;
					}
				}

				while (configLine.find("VertexDisableString") == std::string::npos)
				{
					getline(configFile, configLine);
					// store the injection point
					startPoint = configLine.find("\"", 0);
					endPoint = configLine.find("\"", startPoint + 1);
					m_vertexDisableString = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
					m_vertexDisableString = removeDoubleSlash(m_vertexDisableString);
				}

				while (configLine.find("PixelDisableString") == std::string::npos)
				{
					getline(configFile, configLine);
					// store the injection point
					startPoint = configLine.find("\"", 0);
					endPoint = configLine.find("\"", startPoint + 1);
					m_pixelDisableString = configLine.substr(startPoint + 1, endPoint - startPoint - 1);
					m_pixelDisableString = removeDoubleSlash(m_pixelDisableString);
				}
			}
			//----------------------------------------------
			// Shader Compiler options
			if (configLine.find("[Shader_Compiler_Options]") != std::string::npos)
			{
				size_t startPoint;

				//----------------------------------------------
				//get the next line
				while (configLine.find("EnableCompilerOptions") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("false") != std::string::npos)
					{
						m_exceptShadersEnabled = false;
					}
					else if (configLine.find("true") != std::string::npos)
					{
						m_exceptShadersEnabled = true;
					}
				}

				while (configLine.find("DisableExceptShaderInterval") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("false") != std::string::npos)
					{
						m_disableExceptShaders = false;
					}
					else if (configLine.find("true") != std::string::npos)
					{
						m_disableExceptShaders = true;
					}
				}

				//----------------------------------------------
				//get the next line
				while (configLine.find("ExceptStartShader") == std::string::npos)
				{
					getline(configFile, configLine);
					// store the start shader
					startPoint = configLine.find("= ", 0);
					std::string value = configLine.substr(startPoint + 1);
					m_shaderStart = std::stoul(value, nullptr, 10);
				}

				//----------------------------------------------
				//get the next line
				while (configLine.find("ExceptEndShader") == std::string::npos)
				{
					getline(configFile, configLine);
					// store the end shader
					startPoint = configLine.find("= ", 0);
					std::string value = configLine.substr(startPoint + 1);
					m_shaderEnd = std::stoul(value, nullptr, 10);
				}
				
				while (configLine.find("DisableCurrentShaders") == std::string::npos)
				{
					getline(configFile, configLine);
					if (configLine.find("false") != std::string::npos)
					{
						m_disableCurrentShader = false;
					}
					else if (configLine.find("true") != std::string::npos)
					{
						m_disableCurrentShader = true;
					}
				}
				

				//----------------------------------------------
				//get the next line
				while (configLine.find("ShaderCompilerStartIndex") == std::string::npos)
				{
					getline(configFile, configLine);
					// store the end shader
					startPoint = configLine.find("= ", 0);
					std::string value = configLine.substr(startPoint + 1);
					debugVertexIndex = std::stoul(value, nullptr, 10);
					debugPixelIndex = std::stoul(value, nullptr, 10);
				}

				//----------------------------------------------
			}

			if (configLine.find("[Stereo_Custom_Params]") != std::string::npos)
			{
				// this is needed in order to parse all the except shaders
				while (configLine.find("[End]") == std::string::npos)
				{
					getline(configFile, configLine);
					//----------------------------------------------
					//get the next line
					if (configLine.find("ShaderProgram") != std::string::npos)
					{
						CUSTOM_SHADER_VALUES_T shaderValue;

						// store the injection point
						size_t startPoint = configLine.find("(", 0);
						size_t middlePoint1 = configLine.find(",", startPoint + 1);
						size_t middlePoint2 = configLine.find(",", middlePoint1 + 1);
						size_t middlePoint3 = configLine.find(",", middlePoint2 + 1);
						size_t middlePoint4 = configLine.find(",", middlePoint3 + 1);
						size_t endPoint = configLine.find(")", middlePoint4 + 1);

						shaderValue.m_programId = atoi(configLine.substr(startPoint + 1, middlePoint1 - startPoint - 1).c_str());
						shaderValue.x = (float)atof(configLine.substr(middlePoint1 + 1, middlePoint2 - middlePoint1 - 1).c_str());
						shaderValue.y = (float)atof(configLine.substr(middlePoint2 + 1, middlePoint3 - middlePoint2 - 1).c_str());
						shaderValue.z = (float)atof(configLine.substr(middlePoint3 + 1, middlePoint4 - middlePoint3 - 1).c_str());
						shaderValue.w = (float)atof(configLine.substr(middlePoint4 + 1, endPoint - middlePoint4 - 1).c_str());


						m_customShaderValues.push_back(shaderValue);
						//----------------------------------------------
					}
				}
			}
		}
		configFile.close();
	}

	// Add the shader compiler increments
	if (m_existingShaders.size() != 0)
	{
		for (size_t i = 0; i < m_existingShaders.size(); i++)
		{
			if (debugVertexIndex == (int)m_existingShaders[i].m_programId)
			{
				debugVertexIndex = m_existingShaders[i].m_programId;
				debugPixelIndex = m_existingShaders[i].m_programId;
				break;
			}
			else if (i == (m_existingShaders.size()-1))
			{
				debugVertexIndex = m_existingShaders[0].m_programId;
				debugPixelIndex = m_existingShaders[0].m_programId;
			}
		}
	}
}
///-------------------------------------------------------------------------------------------

std::string ShaderManager::getShaderSource(GLuint shaderId)
{
	// Need to reset the string everytime we get here!!!
	std::string returnSournce = "";

	//The size of the buffer required to store the returned source code string can be obtained by calling glGetShader with the value GL_SHADER_SOURCE_LENGTH.
	GLsizei size = 0;
	(*orig_glGetShader)(shaderId, GL_SHADER_SOURCE_LENGTH, &size);

	// Get the current source
	GLchar* source = new GLchar[size];
	GLsizei returnSize;

	if (orig_glGetShaderSource == 0x00)
	{
		orig_glGetShaderSource = (funct_glGetShaderSource_t)orig_wglGetProcAddress("glGetShaderSource");
	}

	(*orig_glGetShaderSource)(shaderId, size, &returnSize, source);

	// We got the source add it in our string
	returnSournce = source;
	
	// Delete the GLchar
	delete[] source;

	return returnSournce;
}
///-------------------------------------------------------------------------------------------

bool ShaderManager::isShaderType(GLuint _shaderID, GLenum _shaderType)
	{
		// only if we don't have it allready
		if (orig_glGetShader == 0x00)
		{
			orig_glGetShader = (funct_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
		}

		GLint shaderOK = GL_FALSE;
		// Get the Shader Type
		(*orig_glGetShader)(_shaderID, GL_SHADER_TYPE, &shaderOK);

		//if the shader is our type
		if (shaderOK == (GLint)_shaderType)
		{
			return true;
		}
		else
		{
			return false;
		}
	}
///-------------------------------------------------------------------------------------------

// Change the shader Source
void ShaderManager::applyShaderSource(GLuint _shaderID, std::string _shaderSource, const GLint *length)
	{
		// only if we don't have it already
		if (orig_glShaderSource == 0x00)
		{
			orig_glShaderSource = (funct_glShaderSource_t)orig_wglGetProcAddress("glShaderSource");
		}
		const GLchar* source = _shaderSource.c_str();
		(*orig_glShaderSource)(_shaderID, 1, &source, length);
	}
///-------------------------------------------------------------------------------------------

// This is where the magic happens!!!!
// Inject Stereoscopy in all SHADERS.. Hell yeaa!!!
std::string ShaderManager::injectVertexModifications(std::string _inputSource, GLuint programId)
	{
		bool stage1 = false;
		bool stage2 = false;
		std::string newSource = _inputSource;
		size_t charToReplace = 0;
		size_t location = 0;

		// We only need to inject if we didn't do it previously
		// Some programs like to reuse the shaders
		if (_inputSource.find("uniform float g_eye;\nuniform float g_eye_separation;\nuniform float g_convergence;\n") == std::string::npos)
		{
			// Insert stereo
			// We have 4 injection points so go through all of them
			for (size_t currentInjectIndex = 0; currentInjectIndex < MAX_VERTEX_INJECTION_POINTS; currentInjectIndex++)
			{
				location = newSource.find(GetVertexInjectionPoint(currentInjectIndex));
				charToReplace = GetVertexInjectionPoint(currentInjectIndex).length();

				while (location != std::string::npos)
				{
					// Are we inserting or replacing
					if (GetVertexIsInsert(currentInjectIndex))
					{
						if ((location = newSource.find(";", location + 1)) != std::string::npos)
						{
							// We insert the string at this position
							newSource.insert(location + 1, GetVertexStereoInjectString(currentInjectIndex) + GetVertexDisableString());
						}
					}
					// Replace
					else
					{
						newSource.replace(location, charToReplace, GetVertexStereoInjectString(currentInjectIndex) + GetVertexDisableString());
						location += GetVertexStereoInjectString(currentInjectIndex).length() + GetVertexDisableString().length();
					}
					location = newSource.find(GetVertexInjectionPoint(currentInjectIndex), location + GetVertexStereoInjectString(currentInjectIndex).length() + GetVertexDisableString().length());
					stage1 = true;
				}

			}
			//Adding the uniforms
			// We have 3 injection points so go through all of them
			for (size_t currentInjectIndex = 0; currentInjectIndex < MAX_UNIFORM_POINTS; currentInjectIndex++)
			{
				location = newSource.find(GetUniformPoint(currentInjectIndex));

				
				if (location != std::string::npos)
				{
					// Need to count until the end of the current gl_Position
					// And apply the flip afterwards
					for (size_t i = location; i < newSource.size(); i++)
					{
						if (newSource[i] == *("\n"))
						{
							// we reached the end of the current instruction so we break
							break;
						}
						else
						{
							// otherwise keep on counting baby!!!!
							location++;
						}
					}
					// Do the shader gl_Position insertion
					newSource.insert(location + 1, "uniform float g_vertexEnabled;\nuniform float g_eye;\nuniform float g_eye_separation;\nuniform float g_convergence;\nuniform vec4 g_custom_params;\nuniform vec4 g_screeninfo;\n");

					stage2 = true;
				}
				// we inserted so we break;
				if (stage2)
				{
					break;
				}
			}
			if ((!stage1 || !stage2) && (GetStereoFailureInfoState()))
			{
				char result[5000];
				sprintf_s(result, 5000, "Shader Injection FAILED on Shader Program %d.", programId);
				MessageBox(NULL, result, "Stereo Shader Injection", MB_OK);
				add_log(result);
			}
			return newSource;
		}
		else
		{
		
			// We return the same source
			return newSource;
		}
	}
///-------------------------------------------------------------------------------------------
	
// Inject our code in the pixel shader
std::string ShaderManager::injectFragmentModification(std::string _inputSource, GLuint programId)
	{
		bool stage1 = false;
		bool stage2 = false;
		std::string newSource = _inputSource;
		size_t location = 0;
		size_t charToReplace = 0;
		// We only need to inject if we didn't do it previously
		// Some programs like to reuse the shaders
		if (_inputSource.find("uniform float g_eye;\nuniform float g_eye_separation;\nuniform float g_convergence;\n") == std::string::npos)
		{
			// Insert stereo
			// We have 3 injection points so go through all of them
			for (size_t currentInjectIndex = 0; currentInjectIndex < MAX_VERTEX_INJECTION_POINTS; currentInjectIndex++)
			{
				location = newSource.find(GetPixelInjectionPoint(currentInjectIndex));
				charToReplace = GetPixelInjectionPoint(currentInjectIndex).length();

				while (location != std::string::npos)
				{
					// Are we inserting or replacing
					if (GetPixelIsInsert(currentInjectIndex))
					{
						if ((location = newSource.find(";", location + 1)) != std::string::npos)
						{
							// We insert the string at this position
							newSource.insert(location + 1, GetPixelStereoInjectString(currentInjectIndex) + GetPixelDisableString());
						}
					}
					// Replace
					else
					{
						newSource.replace(location, charToReplace, GetPixelStereoInjectString(currentInjectIndex) + GetPixelDisableString());
						location += GetPixelStereoInjectString(currentInjectIndex).length() + GetPixelDisableString().length();
					}
					location = newSource.find(GetPixelInjectionPoint(currentInjectIndex), location + GetPixelStereoInjectString(currentInjectIndex).length() + GetPixelDisableString().length());
					stage1 = true;
				}
			}

			//Adding the uniforms
			// We have 3 injection points so go through all of them
			for (size_t currentInjectIndex = 0; currentInjectIndex < MAX_UNIFORM_POINTS; currentInjectIndex++)
			{
				location = newSource.find(GetUniformPoint(currentInjectIndex));

				if (location != std::string::npos)
				{
					// Need to count until the end of the current gl_Position
					// And apply the flip afterwards
					for (size_t i = location; i < newSource.size(); i++)
					{
						if (newSource[i] == *("\n"))
						{
							// we reached the end of the current instruction so we break
							break;
						}
						else
						{
							// otherwise keep on counting baby!!!!
							location++;
						}
					}
					// Do the shader gl_Position insertion
					newSource.insert(location + 1, "uniform float g_pixelEnabled;\nuniform float g_eye;\nuniform float g_eye_separation;\nuniform float g_convergence;\nuniform vec4 g_custom_params;\nuniform vec4 g_screeninfo;\n");

					stage2 = true;
				}
				// we inserted so we break;
				if (stage2)
				{
					break;
				}
			}
			if ((!stage1 || !stage2) && (GetStereoFailureInfoState()))
			{
				char result[5000];
				sprintf_s(result, 5000, "Pixel Shader Injection FAILED on Shader Program %d.", programId);
				MessageBox(NULL, result, "Pixel Shader Injection", MB_OK);
				add_log(result);
			}
			return newSource;
		}
		else
		{
			// We return the same source
			return newSource;
		}
	}
///-------------------------------------------------------------------------------------------

bool ShaderManager::ApplyExceptionShaders(std::string &newShaderSource, GLenum shaderType, DWORD CRC32)
{
	bool isSet = false;

	// Get all the required functions
	orig_glCreateShader = (func_glCreateShader)orig_wglGetProcAddress("glCreateShader");
	if (orig_glCreateShader == 0x0)
		add_log("glCreateShader not found !!!!");
	orig_glCompileShader = (func_glCompileShader)orig_wglGetProcAddress("glCompileShader");
	if (orig_glCompileShader == 0x0)
		add_log("glCompileShader not found !!!!");
	orig_glGetShaderiv = (func_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
	if (orig_glGetShaderiv == 0x0)
		add_log("glGetShaderiv not found !!!!");
	orig_glGetAttachedShaders = (funct_glGetAttachedShaders_t)orig_wglGetProcAddress("glGetAttachedShaders");
	if (orig_glGetAttachedShaders == 0x0)
		add_log("orig_glGetAttachedShaders not found !!!!");
	orig_glAttachShader = (func_glAttachShader_t)orig_wglGetProcAddress("glAttachShader");
	if (orig_glAttachShader == 0x0)
		add_log("glAttachShader not found !!!!");
	orig_glDetachShader = (func_glDetachShader_t)orig_wglGetProcAddress("glDetachShader");
	if (orig_glDetachShader == 0x0)
		add_log("glDetachShader not found !!!!");
	orig_glDeleteShader = (func_glDeleteShader_t)orig_wglGetProcAddress("glDeleteShader");
	if (orig_glDeleteShader == 0x0)
		add_log("glDeleteShader not found !!!!");


	// check if all the shaders have been applied
	if (m_allShadersApplied == false)
	{
		for (size_t i = 0; i < m_exceptShaders.size(); i++)
		{
			if (m_exceptShaders[i].m_isApplied == false)
			{
				// not all shaders have been applied
				break;
			}
			else if (i == m_exceptShaders.size() - 1)
			{
				m_allShadersApplied = true;
				add_log("All modified shaders have been applied");
			}
		}
	}

	if (m_allShadersApplied == false)
	{
		for (size_t i = 0; i < m_exceptShaders.size(); i++)
		{
			if (
				(CRC32 == m_exceptShaders[i].m_CRC32) &&
				(shaderType == m_exceptShaders[i].m_shaderType) &&
				(m_exceptShaders[i].m_isApplied == false)
				)

			{
				std::string shaderSource;

				// Read our shader source for Vertex
				if (m_exceptShaders[i].m_shaderSourceCode != "")
				{
					// we have a source file
					std::ifstream shaderFile("Shaders/" + m_exceptShaders[i].m_shaderSourceCode);
					std::string shaderLine;

					//file is opened
					if (shaderFile.is_open())
					{
						// read one line at a time
						while (getline(shaderFile, shaderLine))
						{
							shaderSource += shaderLine + "\n";
						}
						shaderFile.close();
						isSet = true;
						newShaderSource = shaderSource;

						// set as applied only if we are not continously searching for new shaders
						if (!IsShaderSwapContinous())
						{
							m_exceptShaders[i].m_isApplied = true;
						}
					}
				}
			}
		}				
	}
	return isSet;
}
///-------------------------------------------------------------------------------------------

void ShaderManager::LoadCustomShaders(void)
{
	m_exceptShaders.resize(0);
	m_exceptPairShaders.resize(0);

	HANDLE hFind;
	WIN32_FIND_DATA data;
	std::string currentFolder = GetPath() + "\\Shaders\\*.glsl";

	// List all the shaders
	hFind = FindFirstFile(currentFolder.c_str(), &data);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		do
		{
			//Decode the filename
			std::stringstream dwordStream;
			DWORD crc32One = 0;
			DWORD crc32Two = 0;
			std::string filename = data.cFileName;
			CUSTOM_SHADER_T newShader;
			CUSTOM_SHADER_PAIR_T newPairShader;
			memset(&newShader, 0x0, sizeof(CUSTOM_SHADER_T));
			newShader.m_shaderSourceCode = filename;

			// Get the Shader Type
			size_t position = filename.find("_");
			std::string temp = filename.substr(position + 1, 1);
				
			if (temp[0] == 'P')
			{
				// Pixel Shader
				newShader.m_shaderType = GL_FRAGMENT_SHADER;
					
			}
			else if (temp[0] == 'V')
			{
				// Vertex Shader
				newShader.m_shaderType = GL_VERTEX_SHADER;
			}

			// Get the 1st crc
			position = filename.find("_", position + 1);
			dwordStream << filename.substr(position + 1);
			dwordStream >> std::hex >> crc32One;
				

			// Get the 2nd crc
			size_t position2 = filename.find("_", position + 1);
			if (position2 != std::string::npos)
			{
				std::stringstream dwordStream2;
				dwordStream2 << filename.substr(position2 + 1);
				dwordStream2 >> std::hex >> crc32Two;
			}

			// Do we have a pair or single shader
			if (crc32One && crc32Two)
			{
				// Save the shader type && code
				newPairShader.m_shaderType = newShader.m_shaderType;
				newPairShader.m_shaderSourceCode = newShader.m_shaderSourceCode;

				if (newShader.m_shaderType == GL_FRAGMENT_SHADER)
				{
					newPairShader.m_CRC32Pixel = crc32One;
					if (crc32Two != 0)
						newPairShader.m_CRC32Vertex = crc32Two;
				}
				else if (newShader.m_shaderType == GL_VERTEX_SHADER)
				{
					newPairShader.m_CRC32Vertex = crc32One;
					if (crc32Two != 0)
						newPairShader.m_CRC32Pixel = crc32Two;
				}

				// Store it in our vector
				// Push the current information in the stack
				newPairShader.m_isApplied = false;
				m_exceptPairShaders.push_back(newPairShader);
			}
			else
			{
				newShader.m_CRC32 = crc32One;

				// Push the current information in the stack
				newShader.m_isApplied = false;
				m_exceptShaders.push_back(newShader);
			}

		} while (FindNextFile(hFind, &data));

		FindClose(hFind);
	}
}
///-------------------------------------------------------------------------------------------

std::string ShaderManager::GetShaderSourceCode(GLuint programId, GLuint startIndex, GLenum shaderType)
{
	std::string shaderSourceCode = "";
	size_t i = startIndex;

	for (; i < (m_existingShaders.size() - 1); i++)
	{
		// Do this only if we don't have an index of zero
		if (i >= 1)
		{
			// Search the previous one
			if ((m_existingShaders[i - 1].m_programId == programId) && (m_existingShaders[i - 1].m_shaderType == shaderType))
			{
				// we found it
				shaderSourceCode = m_existingShaders[i - 1].m_shaderSourceCode;
				break;
			}
		}
		// Search the next one
		if ((m_existingShaders[i + 1].m_programId == programId) && (m_existingShaders[i + 1].m_shaderType == shaderType))
		{
			// we found it
			shaderSourceCode = m_existingShaders[i + 1].m_shaderSourceCode;
			break;
		}
	}
	return shaderSourceCode;
}
///-------------------------------------------------------------------------------------------

// Structure containg the values that we currently send.
extern CUSTOM_SHADER_VALUES_T g_customAltShaderValues;
void ShaderManager::ApplyCustomShaderParams(SHADER_3D_UNIFORMS_T &uniforms)
{
	// Values that are are going to send to the shaders
	float shaderCustomX = -1.0;;
	float shaderCustomY = -1.0;
	float shaderCustomZ = -1.0;
	float shaderCustomW = -1.0;

	for (size_t i = 0; i < m_customShaderValues.size(); i++)
	{
		// -1 is wildcard and params are sent to all shaders
		if ((uniforms.currentProgram == m_customShaderValues[i].m_programId) || (m_customShaderValues[i].m_programId == -1))
		{
			shaderCustomX = m_customShaderValues[i].x;
			shaderCustomY = m_customShaderValues[i].y;
			shaderCustomZ = m_customShaderValues[i].z;
			shaderCustomW = m_customShaderValues[i].w;
			//done and break;
			break;
		}
	}

	// Apply our real-time shader params
	if (g_customAltShaderValues.x != -1.0f)
		shaderCustomX = g_customAltShaderValues.x;
	if (g_customAltShaderValues.y != -1.0f)
		shaderCustomY = g_customAltShaderValues.y;
	if (g_customAltShaderValues.z != -1.0f)
		shaderCustomZ = g_customAltShaderValues.z;
	if (g_customAltShaderValues.w != -1.0f)
		shaderCustomW = g_customAltShaderValues.w;

	// send custom params
	if (uniforms.location_custom_params != -1)
		(*orig_glUniform4f)(uniforms.location_custom_params, shaderCustomX, shaderCustomY, shaderCustomZ, shaderCustomW);
}
///-------------------------------------------------------------------------------------------

void ShaderManager::ApplyCustomShaderParamsARB(SHADER_3D_UNIFORMS_T &uniforms)
{
	for (size_t i = 0; i < m_customShaderValues.size(); i++)
	{
		// -1 is wildcard and params are sent to all shaders
		if ((uniforms.currentProgram == m_customShaderValues[i].m_programId) || (m_customShaderValues[i].m_programId == -1))
		{
			//Set the Separation and convergence
			(*orig_glUniform4fARB)(uniforms.location_custom_params, m_customShaderValues[i].x, m_customShaderValues[i].y, m_customShaderValues[i].z, m_customShaderValues[i].w);
			//done and break;
			break;
		}
	}
}
///-------------------------------------------------------------------------------------------

BOOL ShaderManager::GetCustomShaderParam(GLuint programId, CUSTOM_SHADER_VALUES_T *pCustomShaderParam)
{
	BOOL result = FALSE;

	for (size_t i = 0; i < m_customShaderValues.size(); i++)
	{
		if (programId == m_customShaderValues[i].m_programId)
		{
			pCustomShaderParam->m_programId = m_customShaderValues[i].m_programId;
			pCustomShaderParam->x = m_customShaderValues[i].x;
			pCustomShaderParam->y = m_customShaderValues[i].y;
			pCustomShaderParam->z = m_customShaderValues[i].z;
			pCustomShaderParam->w = m_customShaderValues[i].w;
			result = TRUE;
			break;
		}
	}
	return result;
}
///-------------------------------------------------------------------------------------------

///-------------------------------------------------------------------------------------------
/// PUBLIC FUNCTIONS
///-------------------------------------------------------------------------------------------

void CheckCompileState(GLuint newShaderId, GLuint crc32)
{
	// Only if We have the option to check for the Compile Enabled we want to do it!
	if (g_reader->GetDevelopmentMode())
	{
		char info[255];
		sprintf_s(info, 255, "Shader Compile FAILED: ShaderID:%d - CRC:%x", newShaderId, crc32);

		MessageBox(NULL, info, "", MB_OK);
		GLint maxLength = 0;
		(*orig_glGetShaderiv)(newShaderId, GL_INFO_LOG_LENGTH, &maxLength);

		//The maxLength includes the NULL character
		GLchar *infoLog;
		infoLog = new GLchar[maxLength];
		orig_glGetShaderInfoLog = (func_glGetShaderInfoLog_t)orig_wglGetProcAddress("glGetShaderInfoLog");
		(*orig_glGetShaderInfoLog)(newShaderId, maxLength, &maxLength, infoLog);

		// Get the current source
		// Get the Source Length
		GLsizei sizeSource = 0;
		(*orig_glGetShader)(newShaderId, GL_SHADER_SOURCE_LENGTH, &sizeSource);
		GLchar* source = new GLchar[sizeSource + 5000];
		GLsizei size;
		orig_glGetShaderSource = (funct_glGetShaderSource_t)orig_wglGetProcAddress("glGetShaderSource");
		(*orig_glGetShaderSource)(newShaderId, sizeSource, &size, source);

		strcat_s(source, (sizeSource + 5000), "\n");
		strcat_s(source, (sizeSource + 5000), "\n");
		strcat_s(source, (sizeSource + 5000), "----------------------------------------------------------------------------\n");
		strcat_s(source, (sizeSource + 5000), "Error Output: \n");
		strcat_s(source, (sizeSource + 5000), infoLog);
		strcat_s(source, (sizeSource + 5000), "----------------------------------------------------------------------------\n");

		MessageBox(NULL, source, "Shader Compile Error:", MB_OK);
		add_log("Shader Compile Error\n");

		char crcStr[50];
		std::ostringstream ss;

		ss << std::hex << crc32;
		sprintf_s(crcStr, "ShaderId: %d With CRC32: %s", newShaderId, ss.str().c_str());
		add_log(crcStr);
		add_log(source);
		delete[] source;
	}
}
///-------------------------------------------------------------------------------------------

void CheckLinkState(GLuint programId)
{
	if (g_reader->GetDevelopmentMode())
	{
		GLint maxLength = 0;

		orig_glGetProgramiv = (func_glGetProgramiv_t)orig_wglGetProcAddress("glGetProgramiv");
		(*orig_glGetProgramiv)(programId, GL_INFO_LOG_LENGTH, &maxLength);

		//The maxLength includes the NULL character
		GLchar *infoLog;
		infoLog = new GLchar[maxLength + 1];
		orig_glGetProgramInfoLog = (func_glGetProgramInfoLog_t)orig_wglGetProcAddress("glGetProgramInfoLog");
		(*orig_glGetProgramInfoLog)(programId, maxLength, &maxLength, infoLog);
		char result[5000];
		sprintf_s(result, 5000, "Linking Error on Program %d", programId);
		strcat_s(result, "----------------------------------------------------------------------------\n\n");
		strcat_s(result, infoLog);
		MessageBox(NULL, result, "Linking Error:", MB_OK);
		delete[] infoLog;
	}
}
///-------------------------------------------------------------------------------------------

void DeleteAndDetachShaders(GLuint progId, GLuint shaderId)
{
	//Detach Shaders
	orig_glDetachShader = (func_glDetachShader_t)orig_wglGetProcAddress("glDetachShader");
	orig_glDeleteShader = (func_glDeleteShader_t)orig_wglGetProcAddress("glDeleteShader");

	if (orig_glDetachShader == 0x0)
	{
		MessageBox(NULL, "glDetachShader not found !!!!", "Detach Error:", MB_OK);
	}
	else
	{
		(*orig_glDetachShader)(progId, shaderId);
	}
	//Delete Shaders
	if (orig_glDeleteShader == 0x0)
	{
		MessageBox(NULL, "glDeleteShader not found !!!!", "Delete Error:", MB_OK);
	}
	else
	{
		(*orig_glDeleteShader)(shaderId);
	}
}
///-------------------------------------------------------------------------------------------

void ExportShader(std::string _shaderType, GLuint _program, DWORD crc32, std::string _shaderSource)
{
	if (g_reader->GetDevelopmentMode())
	{
		//Make our dir
		_mkdir("DebugShaders");

		std::ostringstream ss;
		ss << std::hex << crc32;

		//make the file
		std::string shaderFilename = "DebugShaders/" + std::to_string(_program) + "_" + _shaderType + "_" + ss.str() + ".glsl";

		std::ofstream shaderFile;
		shaderFile.open(shaderFilename);
		shaderFile << _shaderSource;
		shaderFile.close();
	}
}
///-------------------------------------------------------------------------------------------

void ShaderManager::ApplyShaderPair(void)
{
	// Get all the required functions
	orig_glCreateShader = (func_glCreateShader)orig_wglGetProcAddress("glCreateShader");
	if (orig_glCreateShader == 0x0)
		add_log("glCreateShader not found !!!!");
	orig_glCompileShader = (func_glCompileShader)orig_wglGetProcAddress("glCompileShader");
	if (orig_glCompileShader == 0x0)
		add_log("glCompileShader not found !!!!");
	orig_glGetShaderiv = (func_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
	if (orig_glGetShaderiv == 0x0)
		add_log("glGetShaderiv not found !!!!");
	orig_glGetAttachedShaders = (funct_glGetAttachedShaders_t)orig_wglGetProcAddress("glGetAttachedShaders");
	if (orig_glGetAttachedShaders == 0x0)
		add_log("orig_glGetAttachedShaders not found !!!!");
	orig_glAttachShader = (func_glAttachShader_t)orig_wglGetProcAddress("glAttachShader");
	if (orig_glAttachShader == 0x0)
		add_log("glAttachShader not found !!!!");
	orig_glDetachShader = (func_glDetachShader_t)orig_wglGetProcAddress("glDetachShader");
	if (orig_glDetachShader == 0x0)
		add_log("glDetachShader not found !!!!");
	orig_glDeleteShader = (func_glDeleteShader_t)orig_wglGetProcAddress("glDeleteShader");
	if (orig_glDeleteShader == 0x0)
		add_log("glDeleteShader not found !!!!");

	// Our indexes
	size_t vertexIndex = 0;
	size_t pixelIndex = 0;
	size_t currentIndex = 0;

	// check if all the shaders have been applied
	if (m_allShaderPairsApplied == false)
	{
		for (size_t i = 0; i < m_exceptPairShaders.size(); i++)
		{
			if (m_exceptPairShaders[i].m_isApplied == false)
			{
				// not all shaders have been applied
				break;
			}
			else if (i == m_exceptPairShaders.size() - 1)
			{
				m_allShaderPairsApplied = true;
				add_log("All modified shaders pairs have been applied");
			}
		}
	}
		
	// check if all the shaders have been applied
	if (m_allShaderPairsApplied == false)
	{
		for (size_t i = 0; i < m_exceptPairShaders.size(); i++)
		{
			while (currentIndex < m_existingShaders.size() - 1)
			{
				// we still have things to do
				// Do we have things left vertex+pixel pair match???
				if ((m_allShaderPairsApplied == false) && (m_exceptPairShaders[i].m_CRC32Vertex != 0) && (m_exceptPairShaders[i].m_CRC32Pixel != 0))
				{
					// Look for a match
					for (size_t i_vertexIndex = currentIndex; i_vertexIndex < m_existingShaders.size(); i_vertexIndex++)
					{
						currentIndex = i_vertexIndex;
						// search for the vertex
						if (m_existingShaders[i_vertexIndex].m_CRC32 == m_exceptPairShaders[i].m_CRC32Vertex)
						{
							// Store our index
							vertexIndex = i_vertexIndex;
							for (size_t i_pixelIndex = 0; i_pixelIndex < m_existingShaders.size(); i_pixelIndex++)
							{
								// search for the pixel and program match!
								if (
									(m_existingShaders[i_pixelIndex].m_CRC32 == m_exceptPairShaders[i].m_CRC32Pixel) &&
									(m_existingShaders[vertexIndex].m_programId == m_existingShaders[i_pixelIndex].m_programId)
									)
								{
									// we found the Pixel ID
									pixelIndex = i_pixelIndex;
									break;
								}
							}
						}
					}

					// WE HAVE A MATCH !!!
					if ((pixelIndex != 0 && vertexIndex != 0) && (!m_exceptPairShaders[i].m_isApplied))
					{
						GLuint currentShaderType = m_exceptPairShaders[i].m_shaderType;
						GLuint pairShaderType = 0;
						int shaderProgramIndex = 0;
						GLint statusOk, LinkOK;
						GLuint newShaderId = 0;
						GLuint pairShaderId = 0;

						bool shaderIsCreated = false;
						bool pairIsCreated = false;
						GLsizei shaderCount;
						GLuint shaders[2] = { 0, 0 };
						(*orig_glGetAttachedShaders)(m_existingShaders[vertexIndex].m_programId, 4, &shaderCount, shaders);

						switch (currentShaderType)
						{
						case GL_VERTEX_SHADER:
							newShaderId = shaders[0];
							pairShaderId = shaders[1];
							shaderProgramIndex = vertexIndex;
							pairShaderType = GL_FRAGMENT_SHADER;
							break;

						case GL_FRAGMENT_SHADER:
							newShaderId = shaders[1];
							pairShaderId = shaders[0];
							shaderProgramIndex = pixelIndex;
							pairShaderType = GL_VERTEX_SHADER;
							break;
						}

						// Generate a new shader
						newShaderId = (*orig_glCreateShader)(m_exceptPairShaders[i].m_shaderType);
						shaderIsCreated = true;

						// Read source file
						std::string newShaderSource;

						// Open the Source File
						std::ifstream shaderFile("Shaders/" + m_exceptPairShaders[i].m_shaderSourceCode);
						std::string shaderLine;

						//file is opened
						if (shaderFile.is_open())
						{
							// read one line at a time
							while (getline(shaderFile, shaderLine))
							{
								newShaderSource += shaderLine + "\n";
							}
							shaderFile.close();
						}


						if (newShaderSource != "")
						{
							// Apply the shader Source
							applyShaderSource(newShaderId, newShaderSource, NULL);

							//Update the Source in the current shader
							m_existingShaders[shaderProgramIndex].m_shaderSourceCode = newShaderSource;
						}
						else
						{
							// Apply the shader Source
							applyShaderSource(newShaderId, m_existingShaders[shaderProgramIndex].m_shaderSourceCode, NULL);
						}

						//Compile vertex shader
						(*orig_glCompileShader)(newShaderId);

						//Test compile
						(*orig_glGetShaderiv)(newShaderId, GL_COMPILE_STATUS, &statusOk);

						if (statusOk == GL_FALSE)
						{
							add_log("ShaderPairSwap() failed!");
							CheckCompileState(newShaderId, m_existingShaders[shaderProgramIndex].m_CRC32);
						}

						//////////////////////////////////////////////////////////////////////////
						/// GET THE SOURCE CODE FOR THE SHADER-PAIR
						//////////////////////////////////////////////////////////////////////////

						// Generate Pair Shader
						pairShaderId = (*orig_glCreateShader)(pairShaderType);
						pairIsCreated = true;

						// Apply the shader Source
						std::string sourceCode = GetShaderSourceCode(m_existingShaders[shaderProgramIndex].m_programId, shaderProgramIndex, pairShaderType);
						applyShaderSource(pairShaderId, sourceCode, NULL);

						// Proceed only if we have source code available at this stage!
						if (sourceCode != "")
						{
							// Attach the New Shader to the Program
							(*orig_glAttachShader)(m_existingShaders[shaderProgramIndex].m_programId, newShaderId);

							//Compile vertex shader
							(*orig_glCompileShader)(pairShaderId);

							//Test compile
							(*orig_glGetShaderiv)(pairShaderId, GL_COMPILE_STATUS, &statusOk);

							if (statusOk == GL_FALSE)
							{
								add_log("ShaderPairSwap() failed!");
								CheckCompileState(pairShaderId, m_existingShaders[shaderProgramIndex].m_CRC32);
							}

							// Attach the PAIR Shader to the Program
							(*orig_glAttachShader)(m_existingShaders[shaderProgramIndex].m_programId, pairShaderId);

							//////////////////////////////////////////////////////////////////////////
							/// AT this stage we have both Shaders Compiled. PS+VS
							/// We can LINK

							// Link the program
							(*orig_glLinkProgram)(m_existingShaders[shaderProgramIndex].m_programId);

							//Check for error
							LinkOK = 0;
							(*orig_glGetProgramiv)(m_existingShaders[shaderProgramIndex].m_programId, GL_LINK_STATUS, &LinkOK);
							if (LinkOK == GL_FALSE)
							{
								CheckLinkState(m_existingShaders[shaderProgramIndex].m_programId);
							}
							else
							{
								// Only if is set to continous mode (Shader Pair)
								if (!IsShaderPairSwapContinous())
									m_exceptPairShaders[i].m_isApplied = true;
							}
						}


						// Clean-up our pair shader
						if (pairIsCreated)
						{
							// Delete our PAIR Shader
							(*orig_glDetachShader)(m_existingShaders[shaderProgramIndex].m_programId, pairShaderId);
							(*orig_glDeleteShader)(pairShaderId);
						}

						// If we created the shader
						// Clean up
						if (shaderIsCreated)
						{
							// We need to clean up here
							(*orig_glDetachShader)(m_existingShaders[shaderProgramIndex].m_programId, newShaderId);
							(*orig_glDeleteShader)(newShaderId);
						}
						// end of if match
					}
				}
				//end for
			}
		}
	}
}
///-------------------------------------------------------------------------------------------
