/*
* Copyright (c) Helifax 2015
*/

// 3D Vision Header
extern "C" {
#include "opengl_3dv.h"
}

// Standard Headers
#include <stdio.h>
#include <direct.h>
#include <iostream>
#include <fstream>
#include <memory>
#include <string>
#include <sstream>
#include <thread>
#include <mutex>
#include <windows.h>

// Our Headers
#include "opengl32.h"
#include "include\ShaderManager.h"
#include "d3d9.h"
#include "NVAPI_367\nvapi.h"
#include "FPSInject/FPSInject.h"
#include "ConfigReader.h"
#include "opengl32_ARB_Variant.h"
#include "wrapper_version.h"
#include "WindowManager.h"
#include "crc32.h"

static bool m_messageBoxShown = false;

// FPSInject
HINSTANCE FPSHandle;
typedef FPSInject* (__cdecl *create)();
typedef void(*destroy)(FPSInject*);
create FPScreate;
destroy FPSDestroy;
static void LoadFPSInject(void);
static bool isFPSInjectHandleValid(void);
static bool m_areExtensionsLoaded = false;
static BOOL g_ProxyLibFixed = FALSE;

// Our object
FPSInject* m_FPSInjector;

// UNITY 5 Uniforms manipulation
// Unity likes to store the uniform locations after shader creation. For some reason if we set our uniforms
// On glUseProgram they become 0 in the shader. Probably since they are overwritten. 
// Thus we need to set our uniforms before we actually draw.
static SHADER_3D_UNIFORMS_T m_uniforms = { -1, -1, -1, -1, -1, -1, -1, 0 };

// Window Information.
// Size, Aspect Ratio and Window Handle.
// Used for sending information to shaders.
// Used for trying to Resize the Window.
unsigned int g_width = 0;
unsigned int g_height = 0;
float g_aspectRatio = 0;
HWND g_MainWindowHandle;
HWND m_originalWindowHandle;
bool g_windowResizeCheck = false;

// 3D Vision
GLD3DBuffers * gl_3dVision_buffers;

//Config reader
extern configReader *g_reader;

// 3D Compatibility Mode - Reshade
static void _3DCompatModeReshadeSwapBuffers(void);
static void _3DCompatModeFixReshadeLib(void);

// Automatic GL_SWAPBUFFERs - Auto hook point
static void _AutomaticGlSwapBuffers(void);

// End frame method used in automatic hooking 
static void _EndFrames(void);

// Used for cycling through the active shaders
int debugVertexIndex = -1;
int debugPixelIndex = -1;
bool printCurrentShader = false;
bool isCurrentShaderVertex = true;
bool isCompileEnable = false;

// Used in shader source SWAP and dumping
static GLuint m_currentGeneratedProgramVertex = 0;
static GLuint m_currentGeneratedProgramPixel = 0;
static GLuint m_currentVertexShaderId = 0;
static GLuint m_currentPixelShaderId = 0;


//********************************************************************************************
// TIME HACK FOR FRAME SYNC
//********************************************************************************************

#ifdef TIME_HACK_ENABLED
void SpeedHack(bool Enabled);
int ExitSpeedHack();

bool SpeedHackEnabled, SpeedHackInitialized;
double TickCount; //Is double so we won't lose data due to rounding, float didn't work for some odd reason
long long PerformanceCount, PerformanceFrequency;
float Acceleration = 1;
int SleepTime = 5; //Updatate frequency for GetTickCount, timeGetTime and QueryPerformanceCounter

typedef DWORD64(__stdcall *timebeginperiod) (UINT Period);
timebeginperiod TimeBeginPeriod = (timebeginperiod)GetProcAddress(LoadLibrary("winmm.dll"), "timeBeginPeriod");

typedef DWORD64(__stdcall *timeendperiod) (UINT Period);
timeendperiod TimeEndPeriod = (timeendperiod)GetProcAddress(LoadLibrary("winmm.dll"), "timeEndPeriod");

void MakePageWritable(DWORD64 Address, DWORD64 ulSize)
{
	MEMORY_BASIC_INFORMATION* mbi = new MEMORY_BASIC_INFORMATION;
	VirtualQuery((void*)Address, mbi, ulSize);
	if (mbi->Protect != PAGE_EXECUTE_READWRITE)
	{
		unsigned long* ulProtect = new unsigned long;
		VirtualProtect((void*)Address, ulSize, PAGE_EXECUTE_READWRITE, ulProtect);
		delete ulProtect;
	}
	delete mbi;
}
///-------------------------------------------------------------------------------------------

bool Jump(DWORD64 Address, void* Function)
{
	__try
	{
		MakePageWritable(Address, 5);
		*(unsigned char*)Address = 0xE9;
		*(DWORD64*)(Address + 1) = ((DWORD64)Function - Address) - 5;
		return true;
	}
	__except (EXCEPTION_EXECUTE_HANDLER) { return false; }
}
///-------------------------------------------------------------------------------------------

void Tick()
{
	TimeBeginPeriod(1);
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
	//_onexit(ExitSpeedHack); //If you get some error saying the stack is corrupt when you exit your program then this is the fault

	float PerformanceSleepTime = (SleepTime * PerformanceFrequency / 1000.0f);
	//int EnableTime = GetTickCount();

	while (true)
	{
		/*double AcceleratedTime = (GetTickCount() - EnableTime) * (Acceleration - 1); //Speedhack inaccuracy testing, rememeber to make sure that you disable the jump at GetTickCount and don't change speed during testing
		int Inaccuracy = (GetTickCount() + AcceleratedTime) - TickCount;*/

		Sleep(SleepTime);
		// If we are enabled
		if (SpeedHackEnabled)
		{
			TickCount += SleepTime * Acceleration;
			PerformanceCount += (long long)(PerformanceSleepTime * Acceleration);
		}
		else
		{
			TickCount += SleepTime;
			PerformanceCount += (long long)(PerformanceSleepTime);
		}
	}
}
///-------------------------------------------------------------------------------------------

int WINAPI NewTickCount()
{
	return (int)TickCount;
}
///-------------------------------------------------------------------------------------------

BOOL WINAPI NewQueryPerformanceCounter(LARGE_INTEGER* Count)
{
	Count->QuadPart = PerformanceCount;
	return TRUE;
}
///-------------------------------------------------------------------------------------------

void SpeedHack(bool Enable)
{
	SpeedHackEnabled = Enable;
	if (Enable == true)
	{
		TickCount = GetTickCount();

		LARGE_INTEGER Ref;
		QueryPerformanceFrequency(&Ref);
		PerformanceFrequency = Ref.QuadPart;

		QueryPerformanceCounter(&Ref);
		PerformanceCount = Ref.QuadPart;

		if (SpeedHackInitialized == false)
		{
			CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&Tick, NULL, 0, NULL);
			SpeedHackInitialized = true;
		}

		void* GetTickCount = GetProcAddress(LoadLibrary("kernel32.dll"), "GetTickCount");
		Jump((DWORD64)GetTickCount, NewTickCount);
		void* TimeGetTime = GetProcAddress(LoadLibrary("winmm.dll"), "timeGetTime");
		Jump((DWORD64)TimeGetTime, NewTickCount);
		void* QueryPerformanceCounter = GetProcAddress(LoadLibrary("kernel32.dll"), "QueryPerformanceCounter");
		Jump((DWORD64)QueryPerformanceCounter, NewQueryPerformanceCounter);
		char temp[255];
		sprintf_s(temp, "Perf addr.: %x\n", QueryPerformanceCounter);
		add_log(temp);
	}
}
///-------------------------------------------------------------------------------------------

int ExitSpeedHack()
{
	TimeEndPeriod(1);
	return 0;
}
#endif

//********************************************************************************************
// END TIME HACK FOR FRAME SYNC
//********************************************************************************************

SHADER_3D_UNIFORMS_T GetCurrent3DUniforms(void)
{
	return m_uniforms;
}

void SetCurrent3DUniforms(SHADER_3D_UNIFORMS_T uniforms)
{
	m_uniforms = uniforms;
}

GLuint GetCurrentGeneratedProgramVertex(void)
{
	return m_currentGeneratedProgramVertex;
}
GLuint GetCurrentGeneratedProgramPixel(void)
{
	return m_currentGeneratedProgramPixel;
}
GLuint GetCurrentVertexShaderId(void)
{
	return m_currentVertexShaderId;
}
GLuint GetCurrentPixelShaderId(void)
{
	return m_currentPixelShaderId;
}

void SetCurrentGeneratedProgramVertex(GLuint value)
{
	m_currentGeneratedProgramVertex = value;
}
void SetCurrentGeneratedProgramPixel(GLuint value)
{
	m_currentGeneratedProgramPixel = value;
}
void SetCurrentVertexShaderId(GLuint value)
{
	m_currentVertexShaderId = value;
}
void SetCurrentPixelShaderId(GLuint value)
{
	m_currentPixelShaderId = value;
}

//********************************************************************************************
// SHADER MANIPULATION
//********************************************************************************************

GLuint sys_glCreateProgram(void)
{
	GLuint program = (*orig_glCreateProgram)();
	return program;
}
///-------------------------------------------------------------------------------------------

// We will inject the stereo here 
void sys_glAttachShader(GLuint program, GLuint shader)
{
	orig_glCompileShader = (func_glCompileShader)orig_wglGetProcAddress("glCompileShader");
	if (orig_glCompileShader == 0x0)
		add_log("glCompileShader not found !!!!");
	orig_glGetShaderiv = (func_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
	if (orig_glGetShaderiv == 0x0)
		add_log("glGetShaderiv not found !!!!");
	orig_glCreateShader = (func_glCreateShader)orig_wglGetProcAddress("glCreateShader");
	if (orig_glCreateShader == 0x0)
		add_log("glCreateShader not found !!!!");


	// Get the correct shader type
	// Get the instance
	ShaderManager *shaderManager = ShaderManager::getInstance();

	// If we have ProgramID embedded in CRC32 set as true in the ini file
	DWORD progCRC32 = 0;
	if (shaderManager->GetProgInCRCState())
	{
		// Add the CRC of the program
		std::string programSring = std::to_string(program);
		progCRC32 = crc32buf(programSring.c_str(), programSring.length());
	}

	// Depends on what type of shader it is we do stuff
	if (shaderManager->isShaderType(shader, GL_VERTEX_SHADER))
	{
		m_currentGeneratedProgramVertex = program;

		//get the original shader Source
		std::string orig_shaderSource = shaderManager->getShaderSource(shader);
		std::string new_shaderSource = orig_shaderSource;

		if (new_shaderSource != "")
		{
			// Calculate the CRC32 before we do any injection
			// Otherwise the CRC32 will change
			DWORD crc32 = crc32buf(orig_shaderSource.c_str(), orig_shaderSource.length());
			crc32 += progCRC32;

			std::ostringstream ss;
			ss << std::hex << crc32;
			add_log("New CRC32 calculated: %s", ss.str().c_str());

			if (shaderManager->GetVertexInjectionState())
			{
				//Apply the custom shaders
				// If we failed apply normal injection
				if (shaderManager->ApplyExceptionShaders(new_shaderSource, GL_VERTEX_SHADER, crc32) == false)
				{
					//Insert Stereo
					new_shaderSource = shaderManager->injectVertexModifications(orig_shaderSource, program);
				}
			}

			// Generate our new Shader
			GLuint newShaderId = (*orig_glCreateShader)(GL_VERTEX_SHADER);
			m_currentVertexShaderId = newShaderId;

			// Swap the Source
			shaderManager->applyShaderSource(newShaderId, new_shaderSource, NULL);

			// Store it as an existing shader
			EXISTING_SHADER_T currentShader;
			currentShader.m_CRC32 = crc32;
			currentShader.m_programId = program;
			currentShader.m_shaderId = shader;
			currentShader.m_shaderSourceCode = new_shaderSource;
			currentShader.m_shaderType = GL_VERTEX_SHADER;
			// Push it back
			shaderManager->addExistingShaderInfo(currentShader);

			// Export the Source
			ExportShader("Vertex", currentShader.m_programId, currentShader.m_CRC32, currentShader.m_shaderSourceCode);
			add_log("New Vertex : %d_ %s.glsl", currentShader.m_programId, ss.str().c_str());

			// We also need to compile the shader before attaching it
			//Compile shader
			(*orig_glCompileShader)(newShaderId);

			//Test compile
			GLint statusOk = 0;
			(*orig_glGetShaderiv)(newShaderId, GL_COMPILE_STATUS, &statusOk);

			if (statusOk == GL_FALSE)
			{
				add_log("glAttachShader() failed on Vertex Shader!");
				CheckCompileState(newShaderId, crc32);
			}

			// We attach after we swapped the sources
			(*orig_glAttachShader)(program, newShaderId);
		}
		else
		{
			// We attach after we swapped the sources
			(*orig_glAttachShader)(program, shader);
		}
	}
	else if (shaderManager->isShaderType(shader, GL_FRAGMENT_SHADER))
	{
		m_currentGeneratedProgramPixel = program;

		//get the original shader Source
		std::string orig_shaderSource = shaderManager->getShaderSource(shader);
		std::string new_shaderSource = orig_shaderSource;

		if (new_shaderSource != "")
		{

			// Calculate the CRC32
			DWORD crc32 = crc32buf(orig_shaderSource.c_str(), orig_shaderSource.length());
			crc32 += progCRC32;

			std::ostringstream ss;
			ss << std::hex << crc32;
			add_log("New CRC32 calculated: %s", ss.str().c_str());

			if (shaderManager->GetVertexInjectionState())
			{
				//Apply the custom shaders
				// If we failed apply normal injection
				if (shaderManager->ApplyExceptionShaders(new_shaderSource, GL_FRAGMENT_SHADER, crc32) == false)
				{
					//Insert Stereo
					new_shaderSource = shaderManager->injectFragmentModification(orig_shaderSource, program);
				}
			}

			// Generate our new Shader
			GLuint newShaderId = (*orig_glCreateShader)(GL_FRAGMENT_SHADER);
			m_currentPixelShaderId = newShaderId;

			// Swap the Source
			shaderManager->applyShaderSource(newShaderId, new_shaderSource, NULL);

			// Store it as an existing shader
			EXISTING_SHADER_T currentShader;
			currentShader.m_CRC32 = crc32;
			currentShader.m_programId = program;
			currentShader.m_shaderId = shader;
			currentShader.m_shaderSourceCode = new_shaderSource;
			currentShader.m_shaderType = GL_FRAGMENT_SHADER;
			// Push it back
			shaderManager->addExistingShaderInfo(currentShader);

			// Export the Source
			ExportShader("Pixel", currentShader.m_programId, currentShader.m_CRC32, currentShader.m_shaderSourceCode);
			add_log("New Pixel : %d_ %s.glsl", currentShader.m_programId, ss.str().c_str());

			//Compile shader
			(*orig_glCompileShader)(newShaderId);

			// We also need to compile the shader before attaching it
			//Test compile
			GLint statusOk = 0;
			(*orig_glGetShaderiv)(newShaderId, GL_COMPILE_STATUS, &statusOk);

			if (statusOk == GL_FALSE)
			{
				add_log("glAttachShader() failed on Pixel Shader!");
				CheckCompileState(newShaderId, crc32);
			}

			// We attach after we swapped the sources
			(*orig_glAttachShader)(program, newShaderId);
		}
		else
		{
			// We attach after we swapped the sources
			(*orig_glAttachShader)(program, shader);
		}
	}
	else
	{
		// We attach after we swapped the sources
		(*orig_glAttachShader)(program, shader);
	}
}
///-------------------------------------------------------------------------------------------

void sys_glShaderSource(GLuint shader, GLsizei count, const GLchar **string, const GLint *length)
{
	// Apply the original source
	(*orig_glShaderSource)(shader, count, string, length);	
}
///-------------------------------------------------------------------------------------------

void sys_glCompileShader(GLuint shader)
{
	// Compile the shader
	(*orig_glCompileShader)(shader);

	// Check to see if the compile was ok
	GLint OK = GL_FALSE;
	orig_glGetShaderiv = (func_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
	(*orig_glGetShaderiv)(shader, GL_COMPILE_STATUS, &OK);

	

	if (OK == GL_FALSE)
	{
		// Get the current source
		// Get the Source Length
		GLsizei sizeSource = 0;
		(*orig_glGetShader)(shader, GL_SHADER_SOURCE_LENGTH, &sizeSource);
		GLchar* source = new GLchar[sizeSource + 5000];
		GLsizei size;
		orig_glGetShaderSource = (funct_glGetShaderSource_t)orig_wglGetProcAddress("glGetShaderSource");
		(*orig_glGetShaderSource)(shader, sizeSource, &size, source);

		// Calculate CRC32
		add_log("glCompileShader() failed!");
		DWORD crc32 = crc32buf(source, sizeSource);
		CheckCompileState(shader, crc32);
	}
}
///-------------------------------------------------------------------------------------------

void sys_glLinkProgram(GLuint program)
{
	// Link the original program now that the correct shaders are swapped
	(*orig_glLinkProgram)(program);

	//Check for error
	GLint LinkOK = 0;
	orig_glGetProgramiv = (func_glGetProgramiv_t)orig_wglGetProcAddress("glGetProgramiv");
	orig_glDetachShader = (func_glDetachShader_t)orig_wglGetProcAddress("glDetachShader");
	orig_glDeleteShader = (func_glDeleteShader_t)orig_wglGetProcAddress("glDeleteShader");
	(*orig_glGetProgramiv)(program, GL_LINK_STATUS, &LinkOK);


	if (LinkOK == GL_FALSE)
	{
		CheckLinkState(program);
	}
	else
	{
		if ((program == m_currentGeneratedProgramPixel) && (program == m_currentGeneratedProgramVertex))
		{
			// Clean up our shaders
			(*orig_glDetachShader)(program, m_currentVertexShaderId);
			(*orig_glDeleteShader)(m_currentVertexShaderId);
			m_currentGeneratedProgramVertex = 0;
			m_currentVertexShaderId = 0;

			(*orig_glDetachShader)(program, m_currentPixelShaderId);
			(*orig_glDeleteShader)(m_currentPixelShaderId);
			m_currentGeneratedProgramPixel = 0;
			m_currentPixelShaderId = 0;
		}
	}

	// Get the correct shader type
	// Get the instance
	ShaderManager *shaderManager = ShaderManager::getInstance();
	// Apply our shader Pairs
	shaderManager->ApplyShaderPair();
}
///-------------------------------------------------------------------------------------------

// add the stereo values
void sys_glUseProgram(GLuint program)
{
	// Get the shaderManager
	ShaderManager * shaderManager = ShaderManager::getInstance();

	if (g_reader->GetDevelopmentMode())
	{
		if (shaderManager->GetVertexInjectionState())
		{
			shaderManager->ApplyDebugExceptionShaders();
		}
	}

	// Use the original program
	(*orig_glUseProgram)(program);

	// Store the current program number
	m_uniforms.currentProgram = program;

	// Get our functions from the driver. Be sure to only take them once
	if (orig_glGetUniformLocation == NULL)
		orig_glGetUniformLocation = (func_glGetUniformLocation_t)(orig_wglGetProcAddress)("glGetUniformLocation");

	//glUniform1f
	if (orig_glUniform1f == NULL)
		orig_glUniform1f = (func_glUniform1f_t)(orig_wglGetProcAddress)("glUniform1f");

	//glUniform4f
	if (!orig_glUniform4f)
		orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");

	if (orig_glGetUniformLocationARB == NULL)
		orig_glGetUniformLocationARB = (func_glGetUniformLocationARB_t)(orig_wglGetProcAddress)("glGetUniformLocationARB");

	//glUniform1f
	if (orig_glUniform1fARB == NULL)
		orig_glUniform1fARB = (func_glUniform1fARB_t)(orig_wglGetProcAddress)("glUniform1fARB");

	//glUniform4f
	if (!orig_glUniform4fARB)
		orig_glUniform4fARB = (func_glUniform4fARB_t)(orig_wglGetProcAddress)("glUniform4fARB");

	if (shaderManager->GetVertexInjectionState())
	{
		// Get the locations of the uniforms
		GLfloat value = 0.0f;
		const GLchar *uniform_eye = (GLchar*)"g_eye";
		const GLchar *uniform_eye_spearation = (GLchar*)"g_eye_separation";
		const GLchar *uniform_convergence = (GLchar*)"g_convergence";
		const GLchar *uniform_vertexEnabled = (GLchar*)"g_vertexEnabled";
		const GLchar *uniform_pixelEnabled = (GLchar*)"g_pixelEnabled";
		const GLchar *uniform_screenInfo = (GLchar*)"g_screeninfo";
		const GLchar *uniform_custom_params = (GLchar*)"g_custom_params";

		m_uniforms.location_eye = (*orig_glGetUniformLocation)(program, uniform_eye);
		m_uniforms.location_eye_separation = (*orig_glGetUniformLocation)(program, uniform_eye_spearation);
		m_uniforms.location_convergence = (*orig_glGetUniformLocation)(program, uniform_convergence);
		m_uniforms.location_vertexEnabled = (*orig_glGetUniformLocation)(program, uniform_vertexEnabled);
		m_uniforms.location_pixelEnabled = (*orig_glGetUniformLocation)(program, uniform_pixelEnabled);
		m_uniforms.location_screenInfo = (*orig_glGetUniformLocation)(program, uniform_screenInfo);
		m_uniforms.location_custom_params = (*orig_glGetUniformLocation)(program, uniform_custom_params);

		(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);
		(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 1.0f);
		if (!orig_glUniform4f)
			orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");
		(*orig_glUniform4f)(m_uniforms.location_screenInfo, (GLfloat)g_width, (GLfloat)g_height, g_aspectRatio, 1.0);

		if (((GLint)program >= 0) && Get3DVisionIsActive())
		{
			if (g_reader->Get3DCompatModeActive() == FALSE)
			{
				//set the uniform inside the shaders
				if (NV3DVisionGetCurrentFrame() == 1)
				{
					//left eye
					// If we aren't the dominant eye
					if (g_reader->GetStereoDominantEye() != DOMINANT_LEFT_EYE)
						value = -1.0f;

					(*orig_glUniform1f)(m_uniforms.location_eye, value);
				}
				else
				{
					//right eye
					// If we aren't the dominant eye
					if (g_reader->GetStereoDominantEye() != DOMINANT_RIGHT_EYE)
						value = 1.0f;
					(*orig_glUniform1f)(m_uniforms.location_eye, value);
				}

				//Set the Separation and convergence
				(*orig_glUniform1f)(m_uniforms.location_eye_separation, Get3DEyeSeparation());
				(*orig_glUniform1f)(m_uniforms.location_convergence, Get3DConvergence());
			}
			else
			{
				(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
				(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
			}
		}
		else if (Get3DVisionIsActive() == FALSE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
			(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
		}
		

		/////////////////////////////////////
		// APPLY CUSTOM SHADER PARAMS
		/////////////////////////////////////
		if (Get3DVisionIsActive())
		{
			shaderManager->ApplyCustomShaderParams(m_uniforms);

			/////////////////////////////////////
			// USED TO DISABLE ALOT OF SHADERS
			/////////////////////////////////////
			if (g_reader->GetDevelopmentMode())
			{
				if (shaderManager->VertexShadersExceptionsEnabled())
				{
					if (((GLint)program >= (GLint)shaderManager->GetExceptionShaderStart() && (GLint)program <= (GLint)shaderManager->GetExceptionShaderEnd()))
					{
						//Set the Separation and convergence
						(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0);
						(*orig_glUniform1f)(m_uniforms.location_convergence, 0);

						// Do we disable the vertex shaders ???
						if (shaderManager->VertexShaderExceptionsDisableShader())
							(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
						else
							(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);

					}
					if (shaderManager->DisableCurrentShader())
					{
						// If we are navigating vertexes
						if (isCurrentShaderVertex)
						{
							// Disable the current Shader
							if (debugVertexIndex == (int)program)
								(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
						}
						else // Pixel Shader
						{
							// Disable the current Shader
							if (debugPixelIndex == (int)program)
								(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 0.0f);
						}
					}
					else
					{
						// If we are navigating vertexes
						if (isCurrentShaderVertex)
						{
							// Disable the current Shader
							if (debugVertexIndex == (int)program)
								(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
						}
						else // Pixel Shader
						{
							// Disable the current Shader
							if (debugPixelIndex == (int)program)
								(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
						}
					}
				}
			}
		}
	}
}

//********************************************************************************************
// END SHADER MANIPULATION
//********************************************************************************************

//********************************************************************************************
// FRAME RENDERING
//********************************************************************************************

// Framebuffer binding
void sys_glBindFramebuffer(GLenum target, GLuint framebuffer)
{
	//Original Frame buffer
	(*orig_glBindFramebuffer)(target, framebuffer);
}
///-------------------------------------------------------------------------------------------

//Initializes and renders
void sys_wglSwapBuffers(HDC hDC)
{
	// Always store the original Handle
	m_originalWindowHandle = WindowFromDC(hDC);

	// If we have the injection point enabled
	if (g_reader->Get3DVisionEnabledStatus())
	{
		// Do we use Reshade Lib?
		if (hProxyLib)
		{
			// Make sure to "patch" Reshade LIB
			_3DCompatModeFixReshadeLib();

			// We only call it if either is generalistic OR is is not generalistic but CM mode is enabled
			if (
				(g_reader->Get3DCompatModeActive() == false && g_reader->GetProxyLibGeneralistic() == true) ||
				(g_reader->Get3DCompatModeActive() == true && g_reader->GetProxyLibGeneralistic() == false) ||
				(g_reader->Get3DCompatModeActive() == true && g_reader->GetProxyLibGeneralistic() == true)
				)
			{
				// Call the original glSwapBuffers that is detoured by Reshade in order to insert the post-process
				(*orig_wglSwapBuffers)(hDC);
			}

			// How do we render ?
			if (g_reader->Get3DCompatModeActive())
			{
				// Copy the content in our DirectX buffers and present
				NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
				NV3DVisionBuffers_activate_right(gl_3dVision_buffers);
				_3DCompatModeReshadeSwapBuffers();
			}
			else
			{
				// If Automatic Hooking point is enabled
				_AutomaticGlSwapBuffers();
			}
		}
		else
		{
			// If Automatic Hooking point is enabled
			_AutomaticGlSwapBuffers();
		}
	}
	else
	{
		(*orig_wglSwapBuffers)(hDC);
	}
}
///-------------------------------------------------------------------------------------------

/// Windows Variant of SwapBuffers
void sys_wglSwapLayerBuffers(HDC hdc, UINT fuPlanes)
{
	// Always store the original Handle
	m_originalWindowHandle = WindowFromDC(hdc);

	// If we have the injection point enabled
	if (g_reader->Get3DVisionEnabledStatus())
	{
		// Do we use Reshade Lib?
		if ((hProxyLib))
		{
			// Make sure to "patch" Reshade LIB
			_3DCompatModeFixReshadeLib();

			// We only call it if either is generalistic OR is is not generalistic but CM mode is enabled
			if (
				(g_reader->Get3DCompatModeActive() == false && g_reader->GetProxyLibGeneralistic() == true) ||
				(g_reader->Get3DCompatModeActive() == true && g_reader->GetProxyLibGeneralistic() == false) ||
				(g_reader->Get3DCompatModeActive() == true && g_reader->GetProxyLibGeneralistic() == true)
				)
			{
				// Call the original glSwapBuffers that is detoured by Reshade in order to insert the post-process
				(*orig_wglSwapBuffers)(hdc);
			}

			// How do we render ?
			if (g_reader->Get3DCompatModeActive())
			{
				// Copy the content in our DirectX buffers and present
				NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
				NV3DVisionBuffers_activate_right(gl_3dVision_buffers);
				_3DCompatModeReshadeSwapBuffers();
			}
			else
			{
				// If Automatic Hooking point is enabled
				_AutomaticGlSwapBuffers();
			}

		}
		else
		{
			// If Automatic Hooking point is enabled
			_AutomaticGlSwapBuffers();
		}
	}
	else
	{
		(*orig_wglSwapLayerBuffers)(hdc, fuPlanes);
	}
}
///-------------------------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////
/// AUTOMATIC SWAPPING
//////////////////////////////////////////////////////////////////////////

// The last Step before actually swapping the buffers
static void _EndFrames(void)
{
	NV3DVisionBuffers_activate_right(gl_3dVision_buffers);

	// Copy the right eye
	if (g_windowResizeCheck == false)
	{
		// We activate our right eye and copy the FBO-Zero content into it!!!!
		NV3DVisionBuffers_copy_right(gl_3dVision_buffers);
	}

	// Deactivate 
	NV3DVisionBuffers_deactivate_right(gl_3dVision_buffers);

	// We rendered everything so first we need to flush.
	NV3DVisionBuffers_flush(gl_3dVision_buffers);

	//check for resize
	if (g_windowResizeCheck == true)
	{
		// Create the new buffers
		g_MainWindowHandle = FindWindowFromProcessId();
		// create DirectX 3D Vision buffers
		NV3DVisionReInitBuffers();
		NV3DVisionBuffers_Dectivate3DVision(gl_3dVision_buffers);
		NV3DVisionBuffers_destroy(gl_3dVision_buffers);

		//Resize complete
		g_windowResizeCheck = false;
		NV3DVisionBuffers_create(gl_3dVision_buffers, g_MainWindowHandle, true);
		NV3DVisionBuffers_Activate3DVision(gl_3dVision_buffers);
	}
	LARGE_INTEGER currentTime;

	QueryPerformanceCounter(&currentTime);
}
///-------------------------------------------------------------------------------------------

//Initializes and renders
static void _AutomaticGlSwapBuffers(void)
{
	//initializes
	if (NV3DVisionIsNotInit())
	{
		// Print just once:)
		add_log("Automatic Hooking Enabled !!!");

		// Load our extensions
		if (!m_areExtensionsLoaded)
		{
			if (ogl_LoadFunctions() == ogl_LOAD_FAILED)
			{
				add_log("Failed to get OpenGL extensions (ogl_LOAD_FAILED)");
			}
			else
			{
				add_log("OpenGL extensions Retrieved !!!");
			}
			m_areExtensionsLoaded = true;
		}

		//Create our buffers
		gl_3dVision_buffers = new GLD3DBuffers;

		g_MainWindowHandle = FindWindowFromProcessId();
		NV3DVisionBuffers_create(gl_3dVision_buffers, g_MainWindowHandle, true);

		// Start our Threads
		StartResizeThread();
		StartNVAPIThread();

		//start the rendering
		NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
		NV3DVisionSetCurrentFrame(1);

		// FPS Inject
		if (isFPSInjectHandleValid())
		{
			m_FPSInjector->synchronizeRendering();
		}
	}
	// If we are initialised
	if (!NV3DVisionIsNotInit())
	{
		if (NV3DVisionGetCurrentFrame() == 2)
		{
			// Run the End Frames Check
			_EndFrames();
			NV3DVisionSetCurrentFrame(1);

			// Sync Left eye
			if (isFPSInjectHandleValid())
				m_FPSInjector->syncLeft();

		}
		else if (NV3DVisionGetCurrentFrame() == 1)
		{
			// Sync Right eye
			if (isFPSInjectHandleValid())
				m_FPSInjector->syncRight();

			if (g_windowResizeCheck == false)
			{
				NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
				NV3DVisionBuffers_copy_left(gl_3dVision_buffers);
				NV3DVisionBuffers_deactivate_left(gl_3dVision_buffers);
			}
			// THIS MIGHT HELP?? OR INTRODUCE EVEN HIGHER EYE-SYNC ISSUES!!!			
			if (g_reader->GetAlternateFrames())
				NV3DVisionBuffers_flush(gl_3dVision_buffers);
			NV3DVisionSetCurrentFrame(2);
		}
	}
}
///-------------------------------------------------------------------------------------------

//********************************************************************************************
// END FRAME RENDERING
//********************************************************************************************

//********************************************************************************************
// RESHADE 3D COMPATIBILITY MODE - SINGLE STEREO PASS
//********************************************************************************************

static void _3DCompatModeFixReshadeLib(void)
{
	// Get the address of glSwapBuffers from Reshade
	// DO THIS JUST ONE TIME
	if (orig_openGl_wglSwapBuffers == 0)
	{
		orig_openGl_wglSwapBuffers = (func_wglSwapBuffers_t)GetProcAddress(hProxyLib, "wglSwapBuffers");
	}

	if (g_ProxyLibFixed == false)
	{
		// Get our PID from the current Process
		DWORD pid = GetPidFromExeName();
		// Open the process and get the process handle
		HANDLE pHandle = OpenProcess(PROCESS_ALL_ACCESS, false, pid);

		// find the address that we want to NOP
		// "Reshade32.dll" + 0x85885 for Reshade 2.03f1 x86
		// "Reshade64.dll" + 0x66C34 for Reshade 2.03f1 x64
		if (strcmp(g_reader->Get3DCompatModeProxyLib().c_str(), "Reshade32.dll") == 0)
		{
			// wglSwapBuffers() + 1E5
			DWORD64 foundAddress = (DWORD64)orig_openGl_wglSwapBuffers + 0x1E5;

			// NOP the CALL/JMP in Reshade.dll
			std::string shellcode = "\x90\x90\x90\x90\x90\x90";
			// Set the page accordingly
			DWORD protect = PAGE_READWRITE;
			VirtualProtect((LPVOID)foundAddress, 1, protect, &protect);

			// Write in Memory
			g_ProxyLibFixed = WriteProcessMemory(pHandle, (LPVOID)foundAddress, shellcode.c_str(), shellcode.length(), NULL);
			VirtualProtect((LPVOID)foundAddress, 1, protect, &protect);
		}
		else  if (strcmp(g_reader->Get3DCompatModeProxyLib().c_str(), "Reshade64.dll") == 0)
		{
			// wglSwapBuffers() + 394
			DWORD64 foundAddress = (DWORD64)orig_openGl_wglSwapBuffers + 0x394;

			// NOP the CALL/JMP in Reshade.dll
			std::string shellcode = "\x90\x90\x90\x90\x90\x90";

			// Set the page accordingly
			DWORD protect = PAGE_READWRITE;
			VirtualProtect((LPVOID)foundAddress, 1, protect, &protect);

			// Write in Memory
			g_ProxyLibFixed = WriteProcessMemory(pHandle, (LPVOID)foundAddress, shellcode.c_str(), shellcode.length(), NULL);
			VirtualProtect((LPVOID)foundAddress, 1, protect, &protect);
		}

		//Close the process handle
		CloseHandle(pHandle);
	}
}
///-------------------------------------------------------------------------------------------

static void _3DCompatModeReshadeSwapBuffers(void)
{
	//initializes
	if (NV3DVisionIsNotInit())
	{
		// Print just once:)
		add_log("Automatic Hooking Enabled !!!");

		// Load our extensions
		if (!m_areExtensionsLoaded)
		{
			if (ogl_LoadFunctions() == ogl_LOAD_FAILED)
			{
				add_log("Failed to get OpenGL extensions (ogl_LOAD_FAILED)");
			}
			else
			{
				add_log("OpenGL extensions Retrieved !!!");
			}
			m_areExtensionsLoaded = true;
		}

		//Create our buffers
		gl_3dVision_buffers = new GLD3DBuffers;

		g_MainWindowHandle = FindWindowFromProcessId();
		NV3DVisionBuffers_create(gl_3dVision_buffers, g_MainWindowHandle, true);
		NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
		NV3DVisionBuffers_activate_right(gl_3dVision_buffers);

		// Start our Threads
		StartResizeThread();
		StartNVAPIThread();
	}
	// If we are initialised
	if (!NV3DVisionIsNotInit())
	{
		if (g_windowResizeCheck == false)
		{
			// Handle left eye
			NV3DVisionBuffers_3DCompat_copy_left(gl_3dVision_buffers);

			// Handle right eye			
			NV3DVisionBuffers_3DCompat_copy_right(gl_3dVision_buffers);
			
			// Present the Frames to the Front-buffer
			NV3DVisionBuffers_flush(gl_3dVision_buffers);
		}

		//check for resize
		if (g_windowResizeCheck == true)
		{
			// Create the new buffers
			g_MainWindowHandle = FindWindowFromProcessId();
			// create DirectX 3D Vision buffers
			NV3DVisionReInitBuffers();
			NV3DVisionBuffers_Dectivate3DVision(gl_3dVision_buffers);
			NV3DVisionBuffers_destroy(gl_3dVision_buffers);

			//Resize complete
			g_windowResizeCheck = false;

			NV3DVisionBuffers_create(gl_3dVision_buffers, g_MainWindowHandle, true);
			NV3DVisionBuffers_Activate3DVision(gl_3dVision_buffers);
			NV3DVisionBuffers_activate_left(gl_3dVision_buffers);
			NV3DVisionBuffers_activate_right(gl_3dVision_buffers);
		}
	}
}
///-------------------------------------------------------------------------------------------

//********************************************************************************************
// END RESHADE 3D COMPATIBILITY MODE - SINGLE STEREO PASS
//********************************************************************************************

//********************************************************************************************
// DRAWING FUNCTIONS
//********************************************************************************************

void sys_glClear(GLbitfield mask)
{
	(*orig_glClear)(mask);
}
///-------------------------------------------------------------------------------------------

void sys_glDrawArrays(GLenum mode, GLint first, GLsizei count)
{
	// Get the shaderManager
	ShaderManager * shaderManager = ShaderManager::getInstance();

	// This will happen only in UNITY 5 Engine
	if (shaderManager->GetUnity5State() && g_reader->Get3DCompatModeActive() == FALSE )
	{
		float value = 0;
		//set the uniform inside the shaders
		if (NV3DVisionGetCurrentFrame() == 1)
		{
			//left eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_LEFT_EYE)
			value = -1.0f;
		}
		else
		{
			// right eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_RIGHT_EYE)
			value = 1.0f;
		}

		const GLchar *uniform_eye = (GLchar*)"g_eye";
		const GLchar *uniform_eye_spearation = (GLchar*)"g_eye_separation";
		const GLchar *uniform_convergence = (GLchar*)"g_convergence";
		const GLchar *uniform_vertexEnabled = (GLchar*)"g_vertexEnabled";
		const GLchar *uniform_pixelEnabled = (GLchar*)"g_pixelEnabled";
		const GLchar *uniform_screenInfo = (GLchar*)"g_screeninfo";
		const GLchar *uniform_custom_params = (GLchar*)"g_custom_params";
		m_uniforms.location_eye = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye);
		m_uniforms.location_eye_separation = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye_spearation);
		m_uniforms.location_convergence = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_convergence);
		m_uniforms.location_vertexEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_vertexEnabled);
		m_uniforms.location_pixelEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_pixelEnabled);
		m_uniforms.location_screenInfo = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_screenInfo);
		m_uniforms.location_custom_params = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_custom_params);

		(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);
		(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 1.0f);

		if (!orig_glUniform4f)
			orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");
		(*orig_glUniform4f)(m_uniforms.location_screenInfo, (GLfloat)g_width, (GLfloat)g_height, g_aspectRatio, 1.0);

		//Set the Separation and convergence
		if (Get3DVisionIsActive() == TRUE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye, value);
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, Get3DEyeSeparation());
			(*orig_glUniform1f)(m_uniforms.location_convergence, Get3DConvergence());
		}
		else if (Get3DVisionIsActive() == FALSE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
			(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
		}

		/////////////////////////////////////
		// APPLY CUSTOM SHADER PARAMS
		/////////////////////////////////////
		shaderManager->ApplyCustomShaderParams(m_uniforms);


		/////////////////////////////////////
		// USED TO DISABLE ALOT OF SHADERS
		/////////////////////////////////////
		if (g_reader->GetDevelopmentMode())
		{
			if (shaderManager->VertexShadersExceptionsEnabled())
			{
				if (
					((GLint)m_uniforms.currentProgram >= (GLint)shaderManager->GetExceptionShaderStart() &&
					(GLint)m_uniforms.currentProgram <= (GLint)shaderManager->GetExceptionShaderEnd())
					)
				{
					//Set the Separation and convergence
					(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0);
					(*orig_glUniform1f)(m_uniforms.location_convergence, 0);

					// Do we disable the vertex shaders ???
					if (shaderManager->VertexShaderExceptionsDisableShader())
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					else
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);

				}
				if (shaderManager->DisableCurrentShader())
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 0.0f);
					}
				}
				else
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
				}
			}
		}
		// end UNITY_5_MODE
	}

	// Call the Drawing function
	(*orig_glDrawArrays) (mode, first, count);
}
///-------------------------------------------------------------------------------------------

void sys_glDrawElements(GLenum mode, GLsizei count, GLenum type, const GLvoid *indices)
{
	// Get the shaderManager
	ShaderManager * shaderManager = ShaderManager::getInstance();

	// This will happen only in UNITY 5 Engine
	if (shaderManager->GetUnity5State() && g_reader->Get3DCompatModeActive() == FALSE)
	{
		float value = 0;
		//set the uniform inside the shaders
		if (NV3DVisionGetCurrentFrame() == 1)
		{
			//left eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_LEFT_EYE)
				value = -1.0f;
		}
		else
		{
			//right eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_RIGHT_EYE)
				value = 1.0f;
		}

		const GLchar *uniform_eye = (GLchar*)"g_eye";
		const GLchar *uniform_eye_spearation = (GLchar*)"g_eye_separation";
		const GLchar *uniform_convergence = (GLchar*)"g_convergence";
		const GLchar *uniform_vertexEnabled = (GLchar*)"g_vertexEnabled";
		const GLchar *uniform_pixelEnabled = (GLchar*)"g_pixelEnabled";
		const GLchar *uniform_screenInfo = (GLchar*)"g_screeninfo";
		const GLchar *uniform_custom_params = (GLchar*)"g_custom_params";
		m_uniforms.location_eye = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye);
		m_uniforms.location_eye_separation = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye_spearation);
		m_uniforms.location_convergence = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_convergence);
		m_uniforms.location_vertexEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_vertexEnabled);
		m_uniforms.location_pixelEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_pixelEnabled);
		m_uniforms.location_screenInfo = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_screenInfo);
		m_uniforms.location_custom_params = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_custom_params);

		(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);
		(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 1.0f);
		if (!orig_glUniform4f)
			orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");
		(*orig_glUniform4f)(m_uniforms.location_screenInfo, (GLfloat)g_width, (GLfloat)g_height, g_aspectRatio, 1.0);

		//Set the Separation and convergence
		if (Get3DVisionIsActive() == TRUE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye, value);
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, Get3DEyeSeparation());
			(*orig_glUniform1f)(m_uniforms.location_convergence, Get3DConvergence());
		}
		else if (Get3DVisionIsActive() == FALSE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
			(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
		}

		/////////////////////////////////////
		// APPLY CUSTOM SHADER PARAMS
		/////////////////////////////////////
		shaderManager->ApplyCustomShaderParams(m_uniforms);

		/////////////////////////////////////
		// USED TO DISABLE ALOT OF SHADERS
		/////////////////////////////////////
		if (g_reader->GetDevelopmentMode())
		{
			if (shaderManager->VertexShadersExceptionsEnabled())
			{
				if (
					((GLint)m_uniforms.currentProgram >= (GLint)shaderManager->GetExceptionShaderStart() &&
					(GLint)m_uniforms.currentProgram <= (GLint)shaderManager->GetExceptionShaderEnd())
					)
				{
					//Set the Separation and convergence
					(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0);
					(*orig_glUniform1f)(m_uniforms.location_convergence, 0);

					// Do we disable the vertex shaders ???
					if (shaderManager->VertexShaderExceptionsDisableShader())
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					else
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);

				}
				if (shaderManager->DisableCurrentShader())
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 0.0f);
					}
				}
				else
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
				}
			}
		}
		// end UNITY_5_MODE
	}

	// Drawing function
	(*orig_glDrawElements) (mode, count, type, indices);
}
///-------------------------------------------------------------------------------------------

void sys_glDrawElementsBaseVertex(GLenum mode, GLsizei count, GLenum type, GLvoid *indices, GLint basevertex)
{
	// Get the shaderManager
	ShaderManager * shaderManager = ShaderManager::getInstance();

	// This will happen only in UNITY 5 Engine
	if (shaderManager->GetUnity5State() && g_reader->Get3DCompatModeActive() == FALSE)
	{
		float value = 0;
		//set the uniform inside the shaders
		if (NV3DVisionGetCurrentFrame() == 1)
		{
			//left eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_LEFT_EYE)
				value = -1.0f;
		}
		else
		{
			//right eye
			// If we aren't the dominant eye
			if (g_reader->GetStereoDominantEye() != DOMINANT_RIGHT_EYE)
				value = 1.0f;
		}

		const GLchar *uniform_eye = (GLchar*)"g_eye";
		const GLchar *uniform_eye_spearation = (GLchar*)"g_eye_separation";
		const GLchar *uniform_convergence = (GLchar*)"g_convergence";
		const GLchar *uniform_vertexEnabled = (GLchar*)"g_vertexEnabled";
		const GLchar *uniform_pixelEnabled = (GLchar*)"g_pixelEnabled";
		const GLchar *uniform_screenInfo = (GLchar*)"g_screeninfo";
		const GLchar *uniform_custom_params = (GLchar*)"g_custom_params";
		m_uniforms.location_eye = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye);
		m_uniforms.location_eye_separation = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_eye_spearation);
		m_uniforms.location_convergence = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_convergence);
		m_uniforms.location_vertexEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_vertexEnabled);
		m_uniforms.location_pixelEnabled = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_pixelEnabled);
		m_uniforms.location_screenInfo = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_screenInfo);
		m_uniforms.location_custom_params = (*orig_glGetUniformLocation)(m_uniforms.currentProgram, uniform_custom_params);

		(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);
		(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 1.0f);
		if (!orig_glUniform4f)
			orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");
		(*orig_glUniform4f)(m_uniforms.location_screenInfo, (GLfloat)g_width, (GLfloat)g_height, g_aspectRatio, 1.0);

		//Set the Separation and convergence
		if (Get3DVisionIsActive() == TRUE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye, value);
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, Get3DEyeSeparation());
			(*orig_glUniform1f)(m_uniforms.location_convergence, Get3DConvergence());
		}
		else if (Get3DVisionIsActive() == FALSE)
		{
			(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
			(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
		}

		/////////////////////////////////////
		// APPLY CUSTOM SHADER PARAMS
		/////////////////////////////////////
		shaderManager->ApplyCustomShaderParams(m_uniforms);

		/////////////////////////////////////
		// USED TO DISABLE ALOT OF SHADERS
		/////////////////////////////////////
		if (g_reader->GetDevelopmentMode())
		{
			if (shaderManager->VertexShadersExceptionsEnabled())
			{
				if (
					((GLint)m_uniforms.currentProgram >= (GLint)shaderManager->GetExceptionShaderStart() &&
					(GLint)m_uniforms.currentProgram <= (GLint)shaderManager->GetExceptionShaderEnd())
					)
				{
					//Set the Separation and convergence
					(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0);
					(*orig_glUniform1f)(m_uniforms.location_convergence, 0);

					// Do we disable the vertex shaders ???
					if (shaderManager->VertexShaderExceptionsDisableShader())
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					else
						(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);

				}
				if (shaderManager->DisableCurrentShader())
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 0.0f);
					}
				}
				else
				{
					// If we are navigating vertexes
					if (isCurrentShaderVertex)
					{
						// Disable the current Shader
						if (debugVertexIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
					else // Pixel Shader
					{
						// Disable the current Shader
						if (debugPixelIndex == (int)m_uniforms.currentProgram)
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					}
				}
			}
		}
		// end UNITY_5_MODE
	}

	// Drawing function
	(*orig_glDrawElementsBaseVertex) (mode, count, type, indices, basevertex);
}
///-------------------------------------------------------------------------------------------

//********************************************************************************************
// END DRAWING FUNCTIONS
//********************************************************************************************

//********************************************************************************************
// MAIN WRAPPER ENTRY POINT & OTHER FUNCTIONS
//********************************************************************************************

BOOL __stdcall DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
{
	(void)lpvReserved;
	(void)hinstDll;

	//Read the config file
	if (g_reader == 0)
		g_reader = new configReader;

	switch (fdwReason)
	{
	case DLL_PROCESS_ATTACH:
		// Init the wrapper
		LoadFPSInject();
		Init();
		// We don't need this as it makes the wrapper not work!!!!
		//DisableThreadLibraryCalls(hOriginalDll);
		break;

	case DLL_PROCESS_DETACH:
		if (hOriginalDll != NULL)
		{
			// Stop the threads
			StopResizeThread();
			StoptNVAPIThread();
			Sleep(500);

			NV3DVisionBuffers_destroy(gl_3dVision_buffers);
			add_log("\nDisabling 3D Vision... Cleaning up and freeing the resources...");
			// Removing the FPS inject
			if (FPSDestroy)
			{
				(FPSDestroy)(m_FPSInjector);
			}
			FreeLibrary(FPSHandle);
			add_log("Application closed successfully !\nEND\n");
			// The the openGL32.dll lib
			FreeLibrary(hOriginalDll);
			hOriginalDll = NULL;

		}
		break;
	}
	return TRUE;
}
#pragma warning(default:4100)
///-------------------------------------------------------------------------------------------

void __cdecl add_log(const char * fmt, ...)
{
	if (g_reader->GetEnableLog())
	{
		const size_t MAX_SIZE = 500000;
		va_list va_alist;
		char logbuf[MAX_SIZE] = "";
		FILE * fp;
		errno_t err;

		va_start(va_alist, fmt);
		_vsnprintf_s(logbuf + strlen(logbuf), MAX_SIZE, sizeof(logbuf)-strlen(logbuf), fmt, va_alist);
		va_end(va_alist);

		err = fopen_s(&fp, "3DVisionWrapper.log", "ab");
		if (err == 0)
		{
			fprintf(fp, "%s", logbuf);
			fprintf(fp, "\n");
			fclose(fp);
		}
	}
}
///-------------------------------------------------------------------------------------------

static void LoadFPSInject(void)
{
	// Load the library
	FPSHandle = LoadLibrary(FPSINJECT_DLL);

	if (FPSHandle != 0)
	{
		// Get the pointers
		FPScreate = (create)GetProcAddress(FPSHandle, "create_object");
		FPSDestroy = (destroy)GetProcAddress(FPSHandle, "destroy_object");

		// Assign the pointer
		m_FPSInjector = (FPScreate)();
	}
}
///-------------------------------------------------------------------------------------------

static bool isFPSInjectHandleValid(void)
{
	bool ret = false;

	if (FPSHandle != 0)
	{
		ret = true;
	}

	return ret;
}
///-------------------------------------------------------------------------------------------

DWORD GetMainThreadId(void)
{
	const std::tr1::shared_ptr<void> hThreadSnapshot(CreateToolhelp32Snapshot(TH32CS_SNAPTHREAD, 0), CloseHandle);
	if (hThreadSnapshot.get() == INVALID_HANDLE_VALUE)
	{
		add_log("GetMainThreadId failed");
		return 0;
	}
	THREADENTRY32 tEntry;
	tEntry.dwSize = sizeof(THREADENTRY32);
	DWORD result = 0;
	DWORD currentPID = GetCurrentProcessId();
	for (BOOL success = Thread32First(hThreadSnapshot.get(), &tEntry);
		!result && success && GetLastError() != ERROR_NO_MORE_FILES;
		success = Thread32Next(hThreadSnapshot.get(), &tEntry))
	{
		if (tEntry.th32OwnerProcessID == currentPID)
		{
			result = tEntry.th32ThreadID;
		}
	}
	return result;
}
///-------------------------------------------------------------------------------------------

PROC sys_wglGetProcAddress(LPCSTR ProcName)
{
	if (!strcmp(ProcName, "glUseProgram"))
	{
		orig_glUseProgram = (funct_glUseProgram_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glUseProgram;
	}
	// ARB Variant
	else if (!strcmp(ProcName, "glUseProgramObjectARB"))
	{
		orig_glUseProgramObjectARB = (funct_glUseProgramObjectARB_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glUseProgramObjectARB;
	}
	else if (!strcmp(ProcName, "glAttachObjectARB"))
	{
		orig_glAttachObjectARB = (func_glAttachShaderARB)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glAttachObjectARB;
	}
	else if (!strcmp(ProcName, "glLinkProgramARB"))
	{
		orig_glLinkProgramARB = (func_glLinkProgramARB)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glLinkProgramARB;
	}
	/*
	// Implementation is MISSING
	// VERY OLD MESA PRE OPENGL 2.0
	// SHOULD NOT BE USED
	else if (!strcmp(ProcName, "glCreateShaderObjectARB"))
	{
		orig_glCreateShaderObjectARB = (func_glCreateShaderObjectARB)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glCreateShaderObjectARB;
	}
	*/
	else if (!strcmp(ProcName, "glCreateProgram"))
	{
		orig_glCreateProgram = (funct_glCreateProgram_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glCreateProgram;
	}
	else if (!strcmp(ProcName, "glAttachShader"))
	{
		orig_glAttachShader = (func_glAttachShader_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glAttachShader;
	}
	else if (!strcmp(ProcName, "glShaderSource"))
	{
		orig_glShaderSource = (funct_glShaderSource_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glShaderSource;
	}
	else if (!strcmp(ProcName, "glBindFramebuffer"))
	{
		orig_glBindFramebuffer = (func_glBindFramebuffer_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glBindFramebuffer;
	}
	else if (!strcmp(ProcName, "glBindFramebufferEXT"))
	{
		orig_glBindFramebuffer = (func_glBindFramebuffer_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glBindFramebuffer;
	}
	else if (!strcmp(ProcName, "glBindFramebufferARB"))
	{
		orig_glBindFramebuffer = (func_glBindFramebuffer_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glBindFramebuffer;
	}
	else if (!strcmp(ProcName, "glLinkProgram"))
	{
		orig_glLinkProgram = (func_glLinkProgram_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glLinkProgram;
	}
	else if (!strcmp(ProcName, "glCompileShader"))
	{
		orig_glCompileShader = (func_glCompileShader)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glCompileShader;
	}
	else if (!strcmp(ProcName, "glDrawElementsBaseVertex"))
	{
		orig_glDrawElementsBaseVertex = (func_glDrawElementsBaseVertex_t)orig_wglGetProcAddress(ProcName);
		return (FARPROC)sys_glDrawElementsBaseVertex;
	}

	// Splash Screen enabled?
	if (g_reader->GetEnableSplashScreen() && !m_messageBoxShown)
	{
		char info[255];
		m_messageBoxShown = true;
		sprintf_s(info, "OpenGL3DVision Wrapper v.%s", OGL3DVISION_VERSION_STR);
		CustomMessageBox(NULL, _T("  \t\t\t\t\t\t               \n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"),
			_T(info), MB_OK);
	}

	return orig_wglGetProcAddress(ProcName);
}
///-------------------------------------------------------------------------------------------

const GLubyte * sys_glGetString(GLenum name)
{
	return (*orig_glGetString) (name);
}
///-------------------------------------------------------------------------------------------