
#include "opengl32_ARB_Variant.h"
#include "crc32.h"
extern configReader *g_reader;
extern GLD3DBuffers * gl_3dVision_buffers;

//////////////////////////////////////////////////////////////////////////
/// ARB VARIANT
//////////////////////////////////////////////////////////////////////////

// Used for cycling through the active shaders
extern int debugVertexIndex;
extern int debugPixelIndex;
extern bool printCurrentShader;
extern bool isCurrentShaderVertex;
extern bool isCompileEnable;
// Window Information.
// Size, Aspect Ratio and Window Handle.
// Used for sending information to shaders.
// Used for trying to Resize the Window.
extern unsigned int g_width;
extern unsigned int g_height;
extern float g_aspectRatio;

// We will inject the stereo here 
void sys_glAttachObjectARB(GLuint program, GLuint shader)
{
	orig_glCompileShaderARB = (func_glCompileShader)orig_wglGetProcAddress("glCompileShaderARB");
	if (orig_glCompileShaderARB == 0x0)
		add_log("glCompileShaderARB not found !!!!");
	orig_glGetShaderiv = (func_glGetShaderiv_t)orig_wglGetProcAddress("glGetShaderiv");
	if (orig_glGetShaderiv == 0x0)
		add_log("glGetShaderiv not found !!!!");
	orig_glCreateShaderObjectARB = (func_glCreateShaderObjectARB)orig_wglGetProcAddress("glCreateShaderObjectARB");
	if (orig_glCreateShaderObjectARB == 0x0)
		add_log("glCreateShaderObjectARB not found !!!!");

	// Get the correct shader type
	// Get the instance
	ShaderManager *shaderManager = ShaderManager::getInstance();

	// If we have ProgramID embedded in CRC32 set as true in the ini file
	DWORD progCRC32 = 0;
	if (shaderManager->GetProgInCRCState())
	{
		// Add the CRC of the program
		std::string programSring = std::to_string(program);
		progCRC32 = crc32buf(programSring.c_str(), programSring.length());
	}

	// Depends on what type of shader it is we do stuff
	if (shaderManager->isShaderType(shader, GL_VERTEX_SHADER))
	{
		SetCurrentGeneratedProgramVertex(program);

		//get the original shader Source
		std::string orig_shaderSource = shaderManager->getShaderSource(shader);
		std::string new_shaderSource = orig_shaderSource;

		if (new_shaderSource != "")
		{
			// Calculate the CRC32 before we do any injection
			// Otherwise the CRC32 will change
			DWORD crc32 = crc32buf(orig_shaderSource.c_str(), orig_shaderSource.length());
			crc32 += progCRC32;

			std::ostringstream ss;
			ss << std::hex << crc32;
			add_log("New CRC32 calculated: %s", ss.str().c_str());

			if (shaderManager->GetVertexInjectionState())
			{
				//Apply the custom shaders
				// If we failed apply normal injection
				if (shaderManager->ApplyExceptionShaders(new_shaderSource, GL_VERTEX_SHADER, crc32) == false)
				{
					//Insert Stereo
					new_shaderSource = shaderManager->injectVertexModifications(orig_shaderSource, program);
				}
			}
			// Generate our new Shader
			GLuint newShaderId = (*orig_glCreateShaderObjectARB)(GL_VERTEX_SHADER);
			SetCurrentVertexShaderId(newShaderId);

			// Swap the Source
			shaderManager->applyShaderSource(newShaderId, new_shaderSource, NULL);

			// Store it as an existing shader
			EXISTING_SHADER_T currentShader;
			currentShader.m_CRC32 = crc32;
			currentShader.m_programId = program;
			currentShader.m_shaderId = shader;
			currentShader.m_shaderSourceCode = new_shaderSource;
			currentShader.m_shaderType = GL_VERTEX_SHADER;
			// Push it back
			shaderManager->addExistingShaderInfo(currentShader);

			// Export the Source
			ExportShader("Vertex", currentShader.m_programId, currentShader.m_CRC32, currentShader.m_shaderSourceCode);
			add_log("New Vertex : %d_ %s.glsl", currentShader.m_programId, ss.str().c_str());

			// We also need to compile the shader before attaching it
			//Compile shader
			(*orig_glCompileShaderARB)(newShaderId);

			//Test compile
			GLint statusOk = 0;
			(*orig_glGetShaderiv)(newShaderId, GL_COMPILE_STATUS, &statusOk);

			if (statusOk == GL_FALSE)
			{
				add_log("glAttachObjectARB() failed on Vertex Shader!");
				CheckCompileState(newShaderId, currentShader.m_CRC32);
			}

			// We attach after we swapped the sources
			(*orig_glAttachObjectARB)(program, newShaderId);
		}
		else
		{
			// We attach after we swapped the sources
			(*orig_glAttachObjectARB)(program, shader);
		}
	}
	else if (shaderManager->isShaderType(shader, GL_FRAGMENT_SHADER))
	{
		SetCurrentGeneratedProgramPixel(program);

		//get the original shader Source
		std::string orig_shaderSource = shaderManager->getShaderSource(shader);
		std::string new_shaderSource = orig_shaderSource;

		if (new_shaderSource != "")
		{
			// Calculate the CRC32
			DWORD crc32 = crc32buf(orig_shaderSource.c_str(), orig_shaderSource.length());
			crc32 += progCRC32;

			if (shaderManager->GetVertexInjectionState())
			{
				//Apply the custom shaders
				// If we failed apply normal injection
				if (shaderManager->ApplyExceptionShaders(new_shaderSource, GL_FRAGMENT_SHADER, crc32) == false)
				{
					//Insert Stereo
					new_shaderSource = shaderManager->injectFragmentModification(orig_shaderSource, program);
				}
			}

			// Generate our new Shader
			GLuint newShaderId = orig_glCreateShaderObjectARB(GL_FRAGMENT_SHADER);
			SetCurrentPixelShaderId(newShaderId);

			// Swap the Source
			shaderManager->applyShaderSource(newShaderId, new_shaderSource, NULL);

			// Store it as an existing shader
			EXISTING_SHADER_T currentShader;
			currentShader.m_CRC32 = crc32;
			currentShader.m_programId = program;
			currentShader.m_shaderId = shader;
			currentShader.m_shaderSourceCode = new_shaderSource;
			currentShader.m_shaderType = GL_FRAGMENT_SHADER;
			// Push it back
			shaderManager->addExistingShaderInfo(currentShader);

			// Export the Source
			ExportShader("Pixel", currentShader.m_programId, currentShader.m_CRC32, currentShader.m_shaderSourceCode);

			//Compile shader
			(*orig_glCompileShaderARB)(newShaderId);

			// We also need to compile the shader before attaching it
			//Test compile
			GLint statusOk = 0;
			(*orig_glGetShaderiv)(newShaderId, GL_COMPILE_STATUS, &statusOk);

			if (statusOk == GL_FALSE)
			{
				add_log("glAttachObjectARB() failed on Pixel Shader!");
				CheckCompileState(newShaderId, currentShader.m_CRC32);
			}

			// We attach after we swapped the sources
			(*orig_glAttachObjectARB)(program, newShaderId);
		}
		else
		{
			// We attach after we swapped the sources
			(*orig_glAttachObjectARB)(program, shader);
		}
	}
	else
	{
		// We attach after we swapped the sources
		(*orig_glAttachObjectARB)(program, shader);
	}
}
///-------------------------------------------------------------------------------------------

void sys_glLinkProgramARB(GLhandleARB program)
{
	// Link the original program now that the correct shaders are swapped
	(orig_glLinkProgramARB)(program);

	// Get all the required functions
	orig_glDetachShader = (func_glDetachShader_t)orig_wglGetProcAddress("glDetachShader");
	if (orig_glDetachShader == 0x0)
		add_log("glDetachShader not found !!!!");
	orig_glDeleteShader = (func_glDeleteShader_t)orig_wglGetProcAddress("glDeleteShader");
	if (orig_glDeleteShader == 0x0)
		add_log("glDeleteShader not found !!!!");

	//Check for error
	GLint LinkOK = 0;
	orig_glGetProgramiv = (func_glGetProgramiv_t)orig_wglGetProcAddress("glGetProgramiv");
	(*orig_glGetProgramiv)(program, GL_LINK_STATUS, &LinkOK);


	if (LinkOK == GL_FALSE)
	{
		CheckLinkState(program);
	}
	else
	{
		if ((program == GetCurrentGeneratedProgramPixel()) && (program == GetCurrentGeneratedProgramVertex()))
		{

			// Clean up our shaders
			(*orig_glDetachShader)(program, GetCurrentVertexShaderId());
			(*orig_glDeleteShader)(GetCurrentVertexShaderId());
			SetCurrentGeneratedProgramVertex(0);
			SetCurrentVertexShaderId(0);

			(*orig_glDetachShader)(program, GetCurrentPixelShaderId());
			(*orig_glDeleteShader)(GetCurrentPixelShaderId());
			SetCurrentGeneratedProgramPixel(0);
			SetCurrentPixelShaderId(0);
		}
	}

	// Get the correct shader type
	// Get the instance
	ShaderManager *shaderManager = ShaderManager::getInstance();
	// Apply our shader Pairs
	shaderManager->ApplyShaderPair();
}
///-------------------------------------------------------------------------------------------

void sys_glUseProgramObjectARB(GLhandleARB program)
{
	SHADER_3D_UNIFORMS_T m_uniforms = GetCurrent3DUniforms();

	// Get the shaderManager
	ShaderManager * shaderManager = ShaderManager::getInstance();

	// UNITY likes to call both the ARB and normal functions
	if (shaderManager->GetUnity5State() == FALSE)
	{
		if (g_reader->GetDevelopmentMode())
		{
			if (shaderManager->GetVertexInjectionState())
			{
				shaderManager->ApplyDebugExceptionShaders();
			}
		}

		// Use the original program
		(*orig_glUseProgramObjectARB)(program);

		// Store the current program number
		m_uniforms.currentProgram = program;

		// Get our functions from the driver. Be sure to only take them once
		if (orig_glGetUniformLocation == NULL)
			orig_glGetUniformLocation = (func_glGetUniformLocation_t)(orig_wglGetProcAddress)("glGetUniformLocation");

		//glUniform1f
		if (orig_glUniform1f == NULL)
			orig_glUniform1f = (func_glUniform1f_t)(orig_wglGetProcAddress)("glUniform1f");

		//glUniform4f
		if (!orig_glUniform4f)
			orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");

		if (orig_glGetUniformLocationARB == NULL)
			orig_glGetUniformLocationARB = (func_glGetUniformLocationARB_t)(orig_wglGetProcAddress)("glGetUniformLocationARB");

		//glUniform1f
		if (orig_glUniform1fARB == NULL)
			orig_glUniform1fARB = (func_glUniform1fARB_t)(orig_wglGetProcAddress)("glUniform1fARB");

		//glUniform4f
		if (!orig_glUniform4fARB)
			orig_glUniform4fARB = (func_glUniform4fARB_t)(orig_wglGetProcAddress)("glUniform4fARB");

		if (shaderManager->GetVertexInjectionState())
		{
			// Get the locations of the uniforms
			GLfloat value = 0.0f;
			const GLchar *uniform_eye = (GLchar*)"g_eye";
			const GLchar *uniform_eye_spearation = (GLchar*)"g_eye_separation";
			const GLchar *uniform_convergence = (GLchar*)"g_convergence";
			const GLchar *uniform_vertexEnabled = (GLchar*)"g_vertexEnabled";
			const GLchar *uniform_pixelEnabled = (GLchar*)"g_pixelEnabled";
			const GLchar *uniform_screenInfo = (GLchar*)"g_screeninfo";
			const GLchar *uniform_custom_params = (GLchar*)"g_custom_params";

			m_uniforms.location_eye = (*orig_glGetUniformLocation)(program, uniform_eye);
			m_uniforms.location_eye_separation = (*orig_glGetUniformLocation)(program, uniform_eye_spearation);
			m_uniforms.location_convergence = (*orig_glGetUniformLocation)(program, uniform_convergence);
			m_uniforms.location_vertexEnabled = (*orig_glGetUniformLocation)(program, uniform_vertexEnabled);
			m_uniforms.location_pixelEnabled = (*orig_glGetUniformLocation)(program, uniform_pixelEnabled);
			m_uniforms.location_screenInfo = (*orig_glGetUniformLocation)(program, uniform_screenInfo);
			m_uniforms.location_custom_params = (*orig_glGetUniformLocation)(program, uniform_custom_params);
			SetCurrent3DUniforms(m_uniforms);

			(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);
			(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 1.0f);
			if (!orig_glUniform4f)
				orig_glUniform4f = (func_glUniform4f_t)(orig_wglGetProcAddress)("glUniform4f");
			(*orig_glUniform4f)(m_uniforms.location_screenInfo, (GLfloat)g_width, (GLfloat)g_height, g_aspectRatio, 1.0);

			if (((GLint)program >= 0) && Get3DVisionIsActive())
			{
				if (g_reader->Get3DCompatModeActive() == FALSE)
				{
					//set the uniform inside the shaders
					if (NV3DVisionGetCurrentFrame() == 1)
					{
						//left eye
						// If we aren't the dominant eye
						if (g_reader->GetStereoDominantEye() != DOMINANT_LEFT_EYE)
							value = -1.0f;

						(*orig_glUniform1f)(m_uniforms.location_eye, value);
					}
					else
					{
						//right eye
						// If we aren't the dominant eye
						if (g_reader->GetStereoDominantEye() != DOMINANT_RIGHT_EYE)
							value = 1.0f;
						(*orig_glUniform1f)(m_uniforms.location_eye, value);
					}

					//Set the Separation and convergence
					(*orig_glUniform1f)(m_uniforms.location_eye_separation, Get3DEyeSeparation());
					(*orig_glUniform1f)(m_uniforms.location_convergence, Get3DConvergence());
				}
				else
				{
					(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
					(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
				}
			}
			else if (Get3DVisionIsActive() == FALSE)
			{
				(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
				(*orig_glUniform1f)(m_uniforms.location_convergence, 0.0f);
			}


			/////////////////////////////////////
			// APPLY CUSTOM SHADER PARAMS
			/////////////////////////////////////
			if (Get3DVisionIsActive())
			{
				shaderManager->ApplyCustomShaderParams(m_uniforms);

				/////////////////////////////////////
				// USED TO DISABLE ALOT OF SHADERS
				/////////////////////////////////////
				if (g_reader->GetDevelopmentMode())
				{
					if (shaderManager->VertexShadersExceptionsEnabled())
					{
						if (((GLint)program >= (GLint)shaderManager->GetExceptionShaderStart() && (GLint)program <= (GLint)shaderManager->GetExceptionShaderEnd()))
						{
							//Set the Separation and convergence
							(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0);
							(*orig_glUniform1f)(m_uniforms.location_convergence, 0);

							// Do we disable the vertex shaders ???
							if (shaderManager->VertexShaderExceptionsDisableShader())
								(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
							else
								(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 1.0f);

						}
						if (shaderManager->DisableCurrentShader())
						{
							// If we are navigating vertexes
							if (isCurrentShaderVertex)
							{
								// Disable the current Shader
								if (debugVertexIndex == (int)program)
									(*orig_glUniform1f)(m_uniforms.location_vertexEnabled, 0.0f);
							}
							else // Pixel Shader
							{
								// Disable the current Shader
								if (debugPixelIndex == (int)program)
									(*orig_glUniform1f)(m_uniforms.location_pixelEnabled, 0.0f);
							}
						}
						else
						{
							// If we are navigating vertexes
							if (isCurrentShaderVertex)
							{
								// Disable the current Shader
								if (debugVertexIndex == (int)program)
									(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
							}
							else // Pixel Shader
							{
								// Disable the current Shader
								if (debugPixelIndex == (int)program)
									(*orig_glUniform1f)(m_uniforms.location_eye_separation, 0.0f);
							}
						}
					}
				}
			}
		}
	}
}
///-------------------------------------------------------------------------------------------


/*
// TO DO:
Add the following wrapped functions if needed in the future:

// Create program
glCreateProgramObjectARB()

// Create the shaders
glCreateShaderObjectARB()

// Shader Source
glShaderSourceARB

// Compile Shader
glCompileShaderARB

*/