#include <iostream>
#include <chrono>
#include <thread>
#include <mutex>


extern "C" {
#include "opengl_3dv.h"
}

#include "opengl32.h"
#include <thread>
#include <VersionHelpers.h>
#include "initguid.h" // this is necessary in order to use IID_IDirect3D9, see: http://support.microsoft.com/kb/130869
#include "d3d9.h"
#include "NVAPI_367\nvapi.h"
#include "wgl_custom.h"
#include "FPSInject/FPSInject.h"
#include "ConfigReader.h"
#include "wrapper_version.h"
#include "WindowManager.h"
#include "ShaderManager.h"
#include "xboxController.h"

// NVAPI 3D VISION
extern GLD3DBuffers * gl_3dVision_buffers;
static NvU8 g_is3DVisionActive = 1;
static float g_separation; 
static float g_convergence;
static float m_currentConvergence;
static float m_currentSeparation;
static float g_eyeSeparation;
static bool m_isCustomSeparationApplied = false;
static bool m_isCustomConvergenceApplied = false;
static bool m_isPressAndHold = false;

// Shader Uniforms Related
CUSTOM_SHADER_VALUES_T g_customAltShaderValues = { 0, -1.0f, -1.0f, -1.0f, -1.0f };

// Mutex 
static std::recursive_mutex m_nvapiMutex;

// Used in init and rendering
static bool m_reInit = false;
static bool m_NV3DVisionIsInInitPhase = true;
static unsigned int m_frameCall = 1;

NV_FRUSTUM_ADJUST_MODE frustum;
StereoHandle nvStereo;

NvAPI_Status status = NVAPI_OK;
BOOL g_NvSurround = FALSE;
BOOL g_NvSLI = FALSE;
static bool m_isResizeCheckThreadInitialised = false;
static bool m_areThreadRunning = true;

// OpenGL-DirectX interoop state
static BOOL g_interop = FALSE;

// NVAPI Polling
static void _NVAPIquery(void);
static void _NVAPi_3DIsActive(void);
static void _DEV_KeyboardKeys(void);

// Config Reader
extern configReader *g_reader;

//Logging
extern void __cdecl add_log(const char * fmt, ...);

// Extern
extern unsigned int g_width;
extern unsigned int g_height;
extern float g_aspectRatio;
extern bool g_windowResizeCheck;
extern BOOL g_fullscreen;

extern int debugVertexIndex;
extern int debugPixelIndex;
extern bool printCurrentShader;
extern bool isCurrentShaderVertex;
extern bool isCompileEnable;
static size_t indexVertex = 0;
static size_t indexPixel = 0;

static unsigned int m_debugKeyCodes[3] = { 0x78, 0x79, 0x7A};
static int m_keyStates[3] = { 0, 0, 0 };
static int m_prevKeyStates[3] = { 0, 0, 0};

// Kill Shader Info Window
static bool KillShaderMessageBox(void);

// Print Current 3D Vision Settings
static void Print3DVisionInfo(void);

// use for hotkeys
static bool IsKeyToggleKeyDown(unsigned int keyCode, int &keyState, int &prevState);

// use for hotkeys
static bool IsAltKeyToggleKeyDown(int keyCodeIndex);
static bool IsKeyDown(int keyCode);
static CXBOXController* g_xController = new CXBOXController(1);
static bool IsXControllerAltKeyToggleKeyDown(int keyCodeIndex);

typedef HRESULT(WINAPI *DIRECT3DCREATE9EXFUNCTION)(UINT SDKVersion, IDirect3D9Ex**);

/*
* NVIDIA stereo
*
* (From nvstereo.h in NVIDIA Direct3D SDK)
*/

typedef struct _Nv_Stereo_Image_Header
{
	unsigned int dwSignature;
	unsigned int dwWidth;
	unsigned int dwHeight;
	unsigned int dwBPP;
	unsigned int dwFlags;
} NVSTEREOIMAGEHEADER, *LPNVSTEREOIMAGEHEADER;

//////////////////////
/// FROM nvstereo.h///
//////////////////////

// dwSignature value
#define NVSTEREO_IMAGE_SIGNATURE	0x4433564e

// dwFlags values
#define SIH_SWAP_EYES				0x00000001
#define SIH_SCALE_TO_FIT			0x00000002
#define SIH_SCALE_TO_FIT2			0x00000004 

//////////////////////
/// FROM WIN SDK 10///
//////////////////////

#define _WIN32_WINNT_WINTHRESHOLD           0x0A00 /* ABRACADABRA_THRESHOLD*/
#define _WIN32_WINNT_WIN10                  0x0A00 /* ABRACADABRA_THRESHOLD*/

// Copy pasted from Windows 10 SDK 
VERSIONHELPERAPI IsWindows10OrGreater()
{
	return IsWindowsVersionOrGreater(HIBYTE(_WIN32_WINNT_WINTHRESHOLD), LOBYTE(_WIN32_WINNT_WINTHRESHOLD), 0);
}
//---------------------------------------------------------------------------->

/*
* Constants
*/
#define DEPTH_STENCIL_FORMAT D3DFMT_D24S8

/*
* Complex stuff!
*
* We need to:
*
* 1. Get the window handle.
* 2. Link the WGL API via the window's device context.
* 3. Make sure we support the WGL NV_DX_interop extension.
* 4. Enable the Direct3D API.
* 5. Create a Direct3D device for our window.
* 5. Enable OpenGL interop for the device.
* 6. Create the Direct3D textures and buffers.
* 7. Create OpenGL textures.
* 8. Bind them to the Direct3D buffers.
* 9. Create a regular OpenGL frame buffer object using the textures.
*
*
* Other useful information:
*  http://developer.download.nvidia.com/opengl/specs/WGL_NV_DX_interop.txt
*  http://sourceforge.net/projects/mingw-w64/forums/forum/723797/topic/5293852
*  http://www.panda3d.org/forums/viewtopic.php?t=11583
*/
///--------------------------------------------------------------------------------------
BOOL NV3DVisionBuffers_create(GLD3DBuffers *gl_d3d_buffers, void *window_handle, bool stereo)
{
	if (m_reInit == false && m_NV3DVisionIsInInitPhase == true)
	{
		add_log("Creating OpenGL/Direct3D bridge...\n");
		ZeroMemory(gl_d3d_buffers, sizeof(GLD3DBuffers));
		gl_d3d_buffers->stereo = stereo;

		HWND hWnd = FindWindowFromProcessId();
		// Invalid handle
		if (!hWnd)
		{
			// Attempt to use the passed handle
			hWnd = (HWND)window_handle;
			if (!hWnd)
			{
				add_log("Cannot retrieve window information... Abort");
				return FALSE;
			}
		}

		WINDOWINFO windowInfo;
		HRESULT result;
		// Fullscreen?
		result = GetWindowInfo(hWnd, &windowInfo);
		if (FAILED(result))
		{
			add_log("Could not get window info");
		}
		
		g_fullscreen = ((windowInfo.dwExStyle & (WS_EX_TOPMOST | WS_POPUP)) != 0);
		if (g_reader->IsFullScreenForced())
		{
			g_fullscreen = TRUE;
		}

		// If we force window mode support
		if (g_reader->IsWindowModeSupportEnabled() == true)
		{
			g_fullscreen = FALSE;
		}
		
		RECT windowRect;
		result = GetClientRect(hWnd, &windowRect);

		unsigned int width = windowRect.right - windowRect.left;
		unsigned int height = windowRect.bottom - windowRect.top;

		if ((width == 0) || (height == 0))
		{
			add_log("Invalid Window Information... Abort");
			return FALSE;
		}

		g_width = width;
		g_height = height;
		g_aspectRatio = (float)width / (float)height;
		gl_d3d_buffers->width = width;
		gl_d3d_buffers->height = height;
		char str[255];
		sprintf_s(str, "Window size: %u, %u", width, height);
		add_log(str);
		
		HDC dc = GetDC(hWnd);
		if (wgl_LoadFunctions(dc) == wgl_LOAD_FAILED)
		{
			add_log("Failed to link to WGL API");
		}

		if (wgl_ext_NV_DX_interop == wgl_LOAD_FAILED)
		{
			add_log("WGL NV_DX_interop extension not supported on this platform");
		}
		add_log("WGL NV_DX_interop extension supported on this platform");

		// Are we a WDDM OS?
		// Higher or equal to vista ?
		BOOL wddmSupport = IsWindowsVistaOrGreater();
		if (wddmSupport)
		{
			add_log("This operating system uses WDDM (it's Windows Vista or later)");
		}
		else {
			add_log("This operating system does not use WDDM (it's prior to Windows Vista). Abort");
		}

		NvU8 stereoSupport = 0;
		// Check and See Stereo Support
		NvAPI_Stereo_IsEnabled(&stereoSupport);

		// Direct3D API
		//
		// (Note: Direct3D/OpenGL interop *requires* the use of 9Ex for WDDM-enabled operating systems)
		IDirect3D9 *d3d;
		// If we have wddmSupport and we have 3D Vision Enabled.
		if (wddmSupport && (stereoSupport == 1))
		{
			// Set Stereo Driver Mode
			NvAPI_Stereo_SetDriverMode(NVAPI_STEREO_DRIVER_MODE_DIRECT);

			// We are dynamically linking to the Direct3DCreate9Ex function, because we
			// don't want to add a run-time dependency for it in our executable, which
			// would make it not run in Windows XP.
			//
			// See: http://msdn.microsoft.com/en-us/library/cc656710.aspx
			
			HMODULE d3dLibrary = LoadLibrary("d3d9.dll");
			if (!d3dLibrary)
			{
				add_log("Failed to link to d3d9.dll");
			}
			gl_d3d_buffers->d3dLibrary = d3dLibrary;

			DIRECT3DCREATE9EXFUNCTION pfnDirect3DCreate9Ex = (DIRECT3DCREATE9EXFUNCTION)GetProcAddress(d3dLibrary, "Direct3DCreate9Ex");
			if (!pfnDirect3DCreate9Ex)
			{
				add_log("Failed to link to Direct3D 9Ex (WDDM)");
			}

			IDirect3D9 *d3dEx;
			result = (*pfnDirect3DCreate9Ex)(D3D_SDK_VERSION, (IDirect3D9Ex**)&d3dEx);
			if (result != D3D_OK)
			{
				add_log("Failed to activate Direct3D 9Ex (WDDM)");
			}

			if (d3dEx->QueryInterface(IID_IDirect3D9, (LPVOID*)&d3d) != S_OK)
			{

				add_log("Failed to cast Direct3D 9Ex to Direct3D 9");
			}

			add_log("Activated Direct3D 9x");

			// Find display mode format
			//
			// (Our buffers will be the same to avoid conversion overhead)
			D3DDISPLAYMODE d3dDisplayMode;
			result = d3d->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3dDisplayMode);
			if (result != D3D_OK)
			{
				add_log("Failed to retrieve adapter display mode");
			}

			// Fixed Transparency issue
			// We need to support Fully ARGB32. Most of the times the adapter returns XRGB32
			// We need to to this only if from this format
			if (d3dDisplayMode.Format == D3DFMT_X8R8G8B8)
			{
				d3dDisplayMode.Format = D3DFMT_A8R8G8B8;
			}

			D3DFORMAT d3dBufferFormat = d3dDisplayMode.Format;

			sprintf_s(str, "Retrieved adapter display mode, format: %u", d3dBufferFormat);
			add_log(str);

			// Make sure devices can support the required formats
			result = d3d->CheckDeviceFormat(
				D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, D3DFMT_X8R8G8B8,
				D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, DEPTH_STENCIL_FORMAT);
			if (result != D3D_OK)
			{
				add_log("Our required formats are not supported");
			}
			add_log("Our required formats are supported");

			// Get the device capabilities
			D3DCAPS9 d3dCaps;
			result = d3d->GetDeviceCaps(
				D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &d3dCaps);
			if (result != D3D_OK)
			{
				add_log("Failed to retrieve device capabilities");
			}

			add_log("Retrieved device capabilities");
			// Verify that we can do hardware vertex processing
			if (d3dCaps.VertexProcessingCaps == 0)
			{
				add_log("Hardware vertex processing is not supported");
			}
			add_log("Hardware vertex processing is supported");

			// Register stereo (must happen *before* device creation)
			if (gl_d3d_buffers->stereo)
			{
				if (NvAPI_Initialize() != NVAPI_OK)
				{
					add_log("Failed to initialize NVAPI");
				}
				else
				{
					add_log("NVAPI initialized");

					NvAPI_Status retVal;
					NvU32 displayID = 0;
					retVal = NvAPI_DISP_GetGDIPrimaryDisplayId(&displayID);

					if (retVal == NVAPI_OK)
					{
						NV_RECT viewports[NV_MOSAIC_MAX_DISPLAYS];
						NvU8 isBezelCorrected;

						retVal = NvAPI_Mosaic_GetDisplayViewportsByResolution(displayID, g_width, g_height, viewports, &isBezelCorrected);

						if (retVal == NVAPI_OK)
						{
							//NVIDIA Surround is enabled.
							add_log("NVIDIA Surround is enabled.");
							g_NvSurround = TRUE;
						}
						else if (retVal == NVAPI_MOSAIC_NOT_ACTIVE)
						{
							// No Surround
							g_NvSurround = FALSE;
							add_log("NVIDIA Single Screen is enabled.");
						}
					}


					// Check for SLI
					NvLogicalGpuHandle *nvGPUHandleLog = new NvLogicalGpuHandle;
					NvPhysicalGpuHandle *nvGPUHandlePhys = new NvPhysicalGpuHandle;

					NvU32 countLogical;
					NvU32 countPhysical;
					NvAPI_EnumLogicalGPUs(nvGPUHandleLog, &countLogical);
					NvAPI_EnumPhysicalGPUs(nvGPUHandlePhys, &countPhysical);

					// If we have more than 1 physical GPU and logical is 1 (SLI enabled)
					if ((countPhysical >= 2) && (countLogical < countPhysical))
					{
						// SLI enabled
						g_NvSLI = TRUE;
						// If we are in fullscreen
						if (g_fullscreen)
						{
							// And Surround is enabled
							if (g_NvSurround)
								add_log("SLI Surround found and enabled.");
							else
								add_log("SLI enabled");
						}
					}
					else // we have a single GPU
					{
						g_NvSLI = FALSE;
						add_log("NVIDIA SLI is Disabled.");
					}

					// Override enable?

				}
			}
			else
			{
				add_log("No stereo Support.");
			}

			D3DPRESENT_PARAMETERS d3dPresent;
			ZeroMemory(&d3dPresent, sizeof(d3dPresent));
			d3dPresent.BackBufferCount = 1;
			d3dPresent.BackBufferFormat = d3dBufferFormat;
			d3dPresent.BackBufferWidth = width;
			d3dPresent.BackBufferHeight = height;
			d3dPresent.SwapEffect = D3DSWAPEFFECT_DISCARD;
			d3dPresent.MultiSampleQuality = 0;
			d3dPresent.MultiSampleType = D3DMULTISAMPLE_NONE;
			d3dPresent.EnableAutoDepthStencil = TRUE;
			d3dPresent.AutoDepthStencilFormat = DEPTH_STENCIL_FORMAT;
			d3dPresent.Flags = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;

			if ((g_NvSLI == FALSE) && (g_NvSurround == FALSE))
			{
				//Single GPU D3DPRESENT_INTERVAL_IMMEDIATE + SWAP EYES
				// SLI enabled but no surround
				if (g_reader->GetSyncToRefreshRate())
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
				else
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			}
			else if ((g_NvSLI == TRUE) && (g_NvSurround == FALSE))
			{
				// SLI enabled but no surround
				if (g_reader->GetSyncToRefreshRate())
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
				else
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			}
			else if ((g_NvSLI == TRUE) && (g_NvSurround == TRUE))
			{
				// Surround and SLI
				if (g_reader->GetSyncToRefreshRate())
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
				else
					d3dPresent.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
			}

			HWND localHwnd = NULL;

			if (g_fullscreen)
			{
				// This only applies in Surround
				d3dPresent.FullScreen_RefreshRateInHz = 120;
				d3dPresent.Windowed = FALSE;
				// Set the correct Handle based on mode !
				localHwnd = hWnd;
				d3dPresent.hDeviceWindow = hWnd; // can be NULL in windowed mode, in which case it will use the arg in CreateDevice

				add_log("Context is in Fullscreen Mode");
			}
			else
			{
				d3dPresent.Windowed = TRUE;
				// Set the correct Handle based on mode !
				localHwnd = hWnd;
				d3dPresent.hDeviceWindow = hWnd; // can be NULL in windowed mode, in which case it will use the arg in CreateDevice
				add_log("Context is in Windowed Mode");
			}
			
			add_log("Windows information for D3D device set.");

			IDirect3DDevice9 *d3dDevice;
			result = d3d->CreateDevice(
				D3DADAPTER_DEFAULT,
				D3DDEVTYPE_HAL,
				localHwnd,
				D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_MULTITHREADED | D3DCREATE_PUREDEVICE,
				&d3dPresent,
				&d3dDevice);
			if (result != D3D_OK)
			{
				switch (result) {
				case D3DERR_DEVICELOST:
					add_log("Failed to create device: device lost");
					break;
				case D3DERR_INVALIDCALL:
					add_log("Failed to create device: invalid call");
					break;
				case D3DERR_NOTAVAILABLE:
					add_log("Failed to create device: not available");
					break;
				case D3DERR_OUTOFVIDEOMEMORY:
					add_log("Failed to create device: out of video memory");
					break;
				default:
					add_log("Failed to create device: unknown error %u", (int)result);
					break;
				}
			}
			gl_d3d_buffers->d3dDevice = d3dDevice;
			add_log("Created device\n");

			// Enable Direct3D/OpenGL interop on device
			HANDLE d3dDeviceInterop = wglDXOpenDeviceNV(d3dDevice);
			if (!d3dDeviceInterop)
			{
				DWORD e = GetLastError();
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				switch (e)
				{
				case ERROR_OPEN_FAILED:
					add_log("Failed to enable OpenGL interop on device");
				case ERROR_NOT_SUPPORTED:
					add_log("Failed to enable OpenGL interop on device: device not supported");
				default:
					add_log("Failed to enable OpenGL interop on device: unknown error %u", (int)e);
					break;
				}
			}
			gl_d3d_buffers->d3dDeviceInterop = d3dDeviceInterop;
			add_log("Enabled OpenGL interop on device");

			// Enable stereo on device (WDDM only)
			if (gl_d3d_buffers->stereo && wddmSupport)
			{
				if (NvAPI_Stereo_CreateHandleFromIUnknown((IUnknown *)d3dDevice, &nvStereo) != NVAPI_OK)
				{
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					add_log("Failed to create NV stereo handle");
				}
				gl_d3d_buffers->nvStereo = nvStereo;
				add_log("Created NV stereo handle on device:");

				if (NvAPI_Stereo_Activate(nvStereo) != NVAPI_OK)
				{
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					add_log("Failed to activate stereo");
				}
				add_log("Activated stereo");
				
				m_nvapiMutex.lock();
				NvAPI_Stereo_GetSeparation(nvStereo, &g_separation);
				NvAPI_Stereo_GetConvergence(nvStereo, &g_convergence);
				NvAPI_Stereo_GetFrustumAdjustMode(nvStereo, &frustum);
				NvAPI_Stereo_GetEyeSeparation(nvStereo, &g_eyeSeparation);
				m_nvapiMutex.unlock();

				// Save default convergence
				m_currentConvergence = g_convergence;
				m_currentSeparation = g_separation;
			}

			// Get device back buffer
			IDirect3DSurface9 *d3dBackBuffer;
			result = d3dDevice->GetBackBuffer(
				0, 0, D3DBACKBUFFER_TYPE_MONO, &d3dBackBuffer);
			if (result != D3D_OK)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to retrieve device back buffer");
			}
			gl_d3d_buffers->d3dBackBuffer = d3dBackBuffer;
			add_log("Retrieved device back buffer");

			// Direct3D textures
			//
			// (Note: we *must* use textures for stereo to show both eyes;
			// a CreateRenderTarget will only work for one eye)

			// Left texture
			IDirect3DTexture9 *d3dLeftColorTexture;
			result = d3dDevice->CreateTexture(
				width, height,
				0, // "levels" (mipmaps)
				0, // usage
				d3dBufferFormat,
				D3DPOOL_DEFAULT,
				&d3dLeftColorTexture,
				NULL);
			if (result != D3D_OK)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to create color texture (left)");
			}
			gl_d3d_buffers->d3dLeftColorTexture = d3dLeftColorTexture;
			add_log("Created color texture(left)");

			IDirect3DTexture9 *d3dRightColorTexture = 0;
			if (gl_d3d_buffers->stereo)
			{
				// Right texture
				result = d3dDevice->CreateTexture(
					width, height,
					0, // "levels" (mipmaps)
					0, // usage
					d3dBufferFormat,
					D3DPOOL_DEFAULT,
					&d3dRightColorTexture,
					NULL);
				if (result != D3D_OK)
				{
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					add_log("Failed to create color texture (right)");
				}
				gl_d3d_buffers->d3dRightColorTexture = d3dRightColorTexture;
				add_log("Created color texture (right)\n", "3D Vision Wrapper");
			}

			// Get Direct3D buffers from textures
			IDirect3DSurface9 *d3dLeftColorBuffer;
			result = d3dLeftColorTexture->GetSurfaceLevel(0, &d3dLeftColorBuffer);
			if (FAILED(result))
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to retrieve color buffer from color texture (left)");
			}
			gl_d3d_buffers->d3dLeftColorBuffer = d3dLeftColorBuffer;
			add_log("Retrieved color buffer from color texture (left)");

			if (gl_d3d_buffers->stereo)
			{
				IDirect3DSurface9 *d3dRightColorBuffer = 0;
				result = d3dRightColorTexture->GetSurfaceLevel(0, &d3dRightColorBuffer);
				if (FAILED(result))
				{
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					add_log("Failed to retrieve color buffer from color texture (right)");
				}
				gl_d3d_buffers->d3dRightColorBuffer = d3dRightColorBuffer;
				add_log("Retrieved color buffer from color texture (right)");
			}

			// Depth/stencil surface
			IDirect3DSurface9 *d3dDepthBuffer;
			result = d3dDevice->CreateDepthStencilSurface(
				width, height,
				DEPTH_STENCIL_FORMAT,
				D3DMULTISAMPLE_NONE, 0,
				FALSE, // Z-buffer discarding
				&d3dDepthBuffer,
				NULL);
			if (result != D3D_OK)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to create depth/stencil surface)");
			}
			gl_d3d_buffers->d3dDepthBuffer = d3dDepthBuffer;
			add_log("Created depth/stencil surface");

			// OpenGL textures
			GLenum error = glGetError();
			glGenTextures(1, &gl_d3d_buffers->texture_left);
			error = glGetError();
			if (error)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to create OpenGL color texture (left)");
			}
			add_log("Created OpenGL color texture (left)");

			if (gl_d3d_buffers->stereo)
			{
				glGenTextures(1, &gl_d3d_buffers->texture_right);
				error = glGetError();
				if (error)
				{
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					add_log("Failed to create OpenGL color texture (right)");
				}
				add_log("Created OpenGL color texture (right)");
			}

			// OpenGL render buffer (one should be enough for stereo?)
			glGenTextures(1, &gl_d3d_buffers->render_buffer);
			error = glGetError();
			if (error)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				add_log("Failed to create OpenGL depth/stencil texture");
			}
			add_log("Created OpenGL depth/stencil texture");

			HANDLE d3dLeftColorInterop = wglDXRegisterObjectNV(d3dDeviceInterop, d3dLeftColorTexture, gl_d3d_buffers->texture_left, GL_TEXTURE_2D, WGL_ACCESS_READ_WRITE_NV);
			if (!d3dLeftColorInterop) 
			{
				DWORD e = GetLastError();
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				switch (e)
				{
				case ERROR_OPEN_FAILED:
					add_log("Failed to bind render target surface to OpenGL texture (left)");
					break;
				case ERROR_INVALID_DATA:
					add_log("Failed to bind render target surface to OpenGL texture (left): invalid data");
					break;
				case ERROR_INVALID_HANDLE:
					add_log("Failed to bind render target surface to OpenGL texture (left): invalid handle");
					break;
				default:
					add_log("Failed to bind render target surface to OpenGL texture (left): unknown error %u", (int)e);
					break;
				}
			}
			gl_d3d_buffers->d3dLeftColorInterop = d3dLeftColorInterop;
			add_log("Bound render target surface to OpenGL texture (left)");

			HANDLE d3dRightColorInterop = wglDXRegisterObjectNV(d3dDeviceInterop, d3dRightColorTexture, gl_d3d_buffers->texture_right, GL_TEXTURE_2D, WGL_ACCESS_READ_WRITE_NV);
			if (gl_d3d_buffers->stereo) {
				if (!d3dRightColorInterop) {
					DWORD e = GetLastError();
					NV3DVisionBuffers_destroy(gl_d3d_buffers);
					switch (e)
					{
					case ERROR_OPEN_FAILED:
						add_log("Failed to bind render target surface to OpenGL texture (right)");
						break;
					case ERROR_INVALID_DATA:
						add_log("Failed to bind render target surface to OpenGL texture (right): invalid data");
						break;
					case ERROR_INVALID_HANDLE:
						add_log("Failed to bind render target surface to OpenGL texture (right): invalid handle");
						break;
					default:
						add_log("Failed to bind render target surface to OpenGL texture (right): unknown error %u", (int)e);
						break;
					}
				}
				gl_d3d_buffers->d3dRightColorInterop = d3dRightColorInterop;
				add_log("Bound render target surface to OpenGL texture (right)");
			}

			// Bind OpenGL render buffer
			HANDLE d3dDepthInterop = wglDXRegisterObjectNV(d3dDeviceInterop, d3dDepthBuffer, gl_d3d_buffers->render_buffer, GL_TEXTURE_2D, WGL_ACCESS_READ_WRITE_NV);
			if (!d3dDepthInterop) {
				DWORD e = GetLastError();
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				switch (e) {
				case ERROR_OPEN_FAILED:
					add_log("Failed to bind depth/stencil surface to OpenGL render buffer");
					break;
				case ERROR_INVALID_DATA:
					add_log("Failed to bind depth/stencil surface to OpenGL render buffer: invalid data");
					break;
				case ERROR_INVALID_HANDLE:
					add_log("Failed to bind depth/stencil surface to OpenGL render buffer: invalid handle");
					break;
				default:
					add_log("Failed to bind depth / stencil surface to OpenGL render buffer : unknown error %u", (int)e);
					break;
				}
			}
			gl_d3d_buffers->d3dDepthInterop = d3dDepthInterop;
			add_log("Bound depth/stencil surface to OpenGL render buffer");

			// Note: if we don't lock the interops first, glCheckFramebufferStatusEXT will fail with an unknown error

			// LOCK ALL 3 HANDLES AT ONCE TO AVOID CPU BOTTLENECK >:< as the GPU will take care of it. Net gain 10fps
			HANDLE gl_handles[3];
			gl_handles[0] = d3dLeftColorInterop;
			gl_handles[1] = d3dRightColorInterop;
			gl_handles[2] = d3dDepthInterop;
			
			// Do the resource LOCK
			wglDXLockObjectsNV(d3dDeviceInterop, 3, gl_handles);

			// OpenGL frame buffer objects
			//
			// Note: why not glFramebufferRenderbufferEXT? Because it's apparently buggy in stereo :/
			// glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, gl_d3d_buffers->render_buffer);

			glGenFramebuffers(1, &gl_d3d_buffers->fbo_left);
			glBindFramebuffer(GL_FRAMEBUFFER, gl_d3d_buffers->fbo_left);
			glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gl_d3d_buffers->texture_left, 0);
			glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, gl_d3d_buffers->render_buffer, 0);
			glFramebufferTexture2D(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_TEXTURE_2D, gl_d3d_buffers->render_buffer, 0);
			GLenum status_left = glCheckFramebufferStatus(GL_FRAMEBUFFER);


			GLenum status_right = 0;
			if (gl_d3d_buffers->stereo)
			{
				glGenFramebuffers(1, &gl_d3d_buffers->fbo_right);
				glBindFramebuffer(GL_FRAMEBUFFER, gl_d3d_buffers->fbo_right);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, gl_d3d_buffers->texture_right, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, gl_d3d_buffers->render_buffer, 0);
				glFramebufferTexture2D(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_TEXTURE_2D, gl_d3d_buffers->render_buffer, 0);
				status_right = glCheckFramebufferStatus(GL_FRAMEBUFFER);

			}

			wglDXUnlockObjectsNV(d3dDeviceInterop, 1, &d3dLeftColorInterop);
			if (gl_d3d_buffers->stereo)
			{
				wglDXUnlockObjectsNV(d3dDeviceInterop, 1, &d3dRightColorInterop);
			}
			wglDXUnlockObjectsNV(d3dDeviceInterop, 1, &d3dDepthInterop);

			// Check for completeness

			if (status_left != GL_FRAMEBUFFER_COMPLETE_EXT)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				switch (status_left)
				{
				case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
					add_log("Failed to create OpenGL frame buffer object (left): attachment");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
					add_log("Failed to create OpenGL frame buffer object (left): missing attachment");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
					add_log("Failed to create OpenGL frame buffer object (left): dimensions");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
					add_log("Failed to create OpenGL frame buffer object (left): formats");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
					add_log("Failed to create OpenGL frame buffer object (left): draw buffer");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
					add_log("Failed to create OpenGL frame buffer object (left): read buffer");
					break;
				case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
					add_log("Failed to create OpenGL frame buffer object (left): unsupported");
					break;
				default:
					add_log("Failed to create OpenGL frame buffer object (left): unknown error %u", (int)status_left);
					break;
				}
			}

			if (status_right != GL_FRAMEBUFFER_COMPLETE_EXT)
			{
				NV3DVisionBuffers_destroy(gl_d3d_buffers);
				switch (status_right)
				{
				case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
					add_log("Failed to create OpenGL frame buffer object (right): attachment");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
					add_log("Failed to create OpenGL frame buffer object (right): missing attachment");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
					add_log("Failed to create OpenGL frame buffer object (right): dimensions");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
					add_log("Failed to create OpenGL frame buffer object (right): formats");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
					add_log("Failed to create OpenGL frame buffer object (right): draw buffer");
					break;
				case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
					add_log("Failed to create OpenGL frame buffer object (right): read buffer");
					break;
				case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
					add_log("Failed to create OpenGL frame buffer object (right): unsupported");
					break;
				default:
					add_log("Failed to create OpenGL frame buffer object (right): unknown error %u", (int)status_left);
					break;
				}
			}

			(*orig_glBindFramebuffer)(GL_FRAMEBUFFER, 0);
			add_log("Created OpenGL frame buffer objects");

			gl_d3d_buffers->initialized = true;
			m_NV3DVisionIsInInitPhase = false;
			add_log("---");
			add_log("3D Vision successfully initialized !");
			add_log("---");
			add_log("Starting Rendering:");
			add_log("---");
		}
		else
		{
			add_log("---");
			add_log("3D Vision is disabled in NVPanel. Please enable it.");
			add_log("---");
			g_reader->Set3DVisionEnabledStatus(false);
		}
		return TRUE;
	}
	return FALSE;
}
///--------------------------------------------------------------------------------------

/* Enable or Disable 3D Vision Mode*/

char NV3DVisionBuffers_Activate3DVision(GLD3DBuffers *gl_d3d_buffers)
{
	NvAPI_Status ret = NVAPI_ERROR;
	if (gl_d3d_buffers)
	{
		ret = NvAPI_Stereo_Activate(gl_d3d_buffers->nvStereo);
	}
	return (char)ret;
}
///--------------------------------------------------------------------------------------

char NV3DVisionBuffers_Dectivate3DVision(GLD3DBuffers *gl_d3d_buffers)
{
	NvAPI_Status ret = NVAPI_ERROR;
	if (gl_d3d_buffers)
	{
		ret = NvAPI_Stereo_Deactivate(gl_d3d_buffers->nvStereo);
	}
	return (char)ret;
}
///--------------------------------------------------------------------------------------

/*
* Before using the OpenGL frame buffer object, we need to make sure to lock
* the Direct3D buffers.
*/
void NV3DVisionBuffers_activate_left(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		BOOL result = FALSE;
		char res[255];

		result = wglDXLockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dLeftColorInterop);
		wglDXLockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dDepthInterop);
		
		if (!result)
		{
			strcpy_s(res, "LEFT EYE: Lock Color Interop: FALSE");
			add_log(res);
		}
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionBuffers_activate_right(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		BOOL result = FALSE;
		char res[255];

		result = wglDXLockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dRightColorInterop);
		wglDXLockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dDepthInterop);
		if (!result)
		{
			strcpy_s(res, "RIGHT EYE: Lock Color Interop: FALSE");
			add_log(res);
		}
	}
}
///--------------------------------------------------------------------------------------

// Copy left image from OGL screen buffer to DX Screen Buffer.
// Perform the flip in one GO
void NV3DVisionBuffers_copy_left(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		// Read to our FBO
		(*orig_glBindFramebuffer)(GL_READ_FRAMEBUFFER, 0);
		(*orig_glBindFramebuffer)(GL_DRAW_FRAMEBUFFER, gl_d3d_buffers->fbo_left);

		glBlitFramebuffer(0, 0, g_width, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
	}
}
///--------------------------------------------------------------------------------------

// Copy right image from OGL screen buffer to DX Screen Buffer.
// Perform the flip in one GO
void NV3DVisionBuffers_copy_right(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		// Read to our FBO
		(*orig_glBindFramebuffer)(GL_READ_FRAMEBUFFER, 0);
		(*orig_glBindFramebuffer)(GL_DRAW_FRAMEBUFFER, gl_d3d_buffers->fbo_right);

		glBlitFramebuffer(0, 0, g_width, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
	}
}
///--------------------------------------------------------------------------------------

// Copy left image from OGL screen buffer to DX Screen Buffer.
// Perform the flip in one GO
// We need to take into account that the image is presented SBS so we only need to copy half of it
void NV3DVisionBuffers_3DCompat_copy_left(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		// Read to our FBO
		(*orig_glBindFramebuffer)(GL_READ_FRAMEBUFFER, 0);
		(*orig_glBindFramebuffer)(GL_DRAW_FRAMEBUFFER, gl_d3d_buffers->fbo_left);
		glBlitFramebuffer(0, 0, g_width/2, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
		//glBlitFramebuffer(0, 0, g_width, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
	}
}
///--------------------------------------------------------------------------------------

// Copy right image from OGL screen buffer to DX Screen Buffer.
// Perform the flip in one GO
// We need to take into account that the image is presented SBS so we only need to copy half of it
void NV3DVisionBuffers_3DCompat_copy_right(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		// Read to our FBO
		(*orig_glBindFramebuffer)(GL_READ_FRAMEBUFFER, 0);
		(*orig_glBindFramebuffer)(GL_DRAW_FRAMEBUFFER, gl_d3d_buffers->fbo_right);

		// Definition
		glBlitFramebuffer(g_width/2, 0, g_width, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
		//glBlitFramebuffer(0, 0, g_width, g_height, 0, g_height, g_width, 0, GL_COLOR_BUFFER_BIT, GL_LINEAR);
	}
}
///--------------------------------------------------------------------------------------

// Clear the LEFT Framebuffer
void NV3DVisionBuffers_clear_left(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER, gl_d3d_buffers->fbo_left);
		(*orig_glClear)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER_EXT, 0);
	}
}
///--------------------------------------------------------------------------------------

// Clear the RIGHT Framebuffer
void NV3DVisionBuffers_clear_right(GLD3DBuffers *gl_d3d_buffers)
{
	if (!m_NV3DVisionIsInInitPhase)
	{
		(*orig_glBindFramebuffer)(GL_DRAW_FRAMEBUFFER, gl_d3d_buffers->fbo_right);
		(*orig_glClear)(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER_EXT, 0);
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionBuffers_deactivate_left(GLD3DBuffers *gl_d3d_buffers)
{
	if ((!m_NV3DVisionIsInInitPhase) || (m_reInit))
	{
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER_EXT, 0);
		BOOL ok = wglDXUnlockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dLeftColorInterop);
		wglDXUnlockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dDepthInterop);

		if (ok != TRUE)
		{
			char res[255];
			strcpy_s(res, "Deactivate LEFT Buffer Interop: FALSE");
			add_log(res);

		}
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionBuffers_deactivate_right(GLD3DBuffers *gl_d3d_buffers)
{
	if ((!m_NV3DVisionIsInInitPhase) || (m_reInit))
	{
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER_EXT, 0);
		BOOL ok = wglDXUnlockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dRightColorInterop);
		wglDXUnlockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 1, &gl_d3d_buffers->d3dDepthInterop);
		
		if (ok != TRUE)
		{
			char res[255];
			strcpy_s(res, "Deactivate RIGHT Buffer Interop: FALSE");
			add_log(res);
			
		}
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionBuffers_deactivate(GLD3DBuffers *gl_d3d_buffers)
{
	(void)gl_d3d_buffers;
	if ((!m_NV3DVisionIsInInitPhase) || (m_reInit))
	{
		HANDLE gl_handles[3];
		BOOL result = FALSE;

		gl_handles[0] = gl_d3d_buffers->d3dLeftColorInterop;
		gl_handles[1] = gl_d3d_buffers->d3dRightColorInterop;
		gl_handles[2] = gl_d3d_buffers->d3dDepthInterop;
		(*orig_glBindFramebuffer)(GL_FRAMEBUFFER_EXT, 0);

		result = wglDXUnlockObjectsNV(gl_d3d_buffers->d3dDeviceInterop, 3, gl_handles);
	}
}
///--------------------------------------------------------------------------------------

/*
* Flushing:
*
* 1. Copy our color buffer to the device's back buffer.
* 2. Present the device back buffer.
*/
void NV3DVisionBuffers_flush(GLD3DBuffers *gl_d3d_buffers)
{
	// Only if we are initialized
	if ((!m_NV3DVisionIsInInitPhase) || (m_reInit))
	{
	IDirect3DDevice9 *d3dDevice = (IDirect3DDevice9 *)gl_d3d_buffers->d3dDevice;
	IDirect3DSurface9 *d3dBackBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dBackBuffer;
	IDirect3DSurface9 *d3dLeftColorBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dLeftColorBuffer;
	IDirect3DSurface9 *d3dRightColorBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dRightColorBuffer;

	HRESULT result;

	// Is 3D Vision Enabled & Activated?
	if ((gl_d3d_buffers->stereo) && (Get3DVisionIsActive() == TRUE))
	{
		// Stretch left color buffer to left side
		NvAPI_Stereo_SetActiveEye(nvStereo, NVAPI_STEREO_EYE_LEFT);
		result = d3dDevice->StretchRect(d3dLeftColorBuffer, NULL, d3dBackBuffer, NULL, D3DTEXF_NONE);


		// Stretch right color buffer to right side
		NvAPI_Stereo_SetActiveEye(nvStereo, NVAPI_STEREO_EYE_RIGHT);
		result = d3dDevice->StretchRect(d3dRightColorBuffer, NULL, d3dBackBuffer, NULL, D3DTEXF_NONE);
	}
	else
	{
		// Stretch left color buffer to back buffer
		result = d3dDevice->StretchRect(d3dLeftColorBuffer, NULL, d3dBackBuffer, NULL, D3DTEXF_NONE);
	}

	// Present back buffer
	result = d3dDevice->Present(NULL, NULL, NULL, NULL);
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionBuffers_destroy(GLD3DBuffers *gl_d3d_buffers) 
{
	// Only if we are allowed to destroy
	if (m_reInit == true && m_NV3DVisionIsInInitPhase == false)
	{
		add_log("Disabling 3D Vision and cleaning up !");

		// First let's deactivate all the buffers
		NV3DVisionBuffers_flush(gl_d3d_buffers);
		NV3DVisionBuffers_deactivate_left(gl_d3d_buffers);
		NV3DVisionBuffers_deactivate_right(gl_d3d_buffers);
		// Be sure that everything is deactivated
		NV3DVisionBuffers_deactivate(gl_d3d_buffers);

		// Reset the frames flags
		m_frameCall = 0;
		//////////////////////////////////////////////////////////////////////////
		/// Destroy the objects
		//////////////////////////////////////////////////////////////////////////

		// Destroy NV stereo handle
		if (gl_d3d_buffers->nvStereo)
		{
			NvAPI_Stereo_DestroyHandle(gl_d3d_buffers->nvStereo);
			gl_d3d_buffers->nvStereo = NULL;
			add_log("Deleting the Stereo Handle");
		}

		// Unbind OpenGL textures from Direct3D buffers
		if (gl_d3d_buffers->d3dDeviceInterop && gl_d3d_buffers->d3dLeftColorInterop)
		{
			if (wglDXUnregisterObjectNV(gl_d3d_buffers->d3dDeviceInterop, gl_d3d_buffers->d3dLeftColorInterop))
			{
				gl_d3d_buffers->d3dLeftColorInterop = NULL;
				add_log("De-registered Left Color buffer");
			}
		}

		if (gl_d3d_buffers->d3dDeviceInterop && gl_d3d_buffers->d3dRightColorInterop)
		{
			wglDXUnregisterObjectNV(gl_d3d_buffers->d3dDeviceInterop, gl_d3d_buffers->d3dRightColorInterop);
			gl_d3d_buffers->d3dRightColorInterop = NULL;
			add_log("De-registered Right Color buffer");
		}

		if (gl_d3d_buffers->d3dDeviceInterop && gl_d3d_buffers->d3dDepthInterop)
		{
			wglDXUnregisterObjectNV(gl_d3d_buffers->d3dDeviceInterop, gl_d3d_buffers->d3dDepthInterop);
			gl_d3d_buffers->d3dDepthInterop = NULL;
			add_log("De-registered Depth buffer");
		}

		// Disable Direct3D/OpenGL interop
		if (gl_d3d_buffers->d3dDeviceInterop)
		{
			wglDXCloseDeviceNV(gl_d3d_buffers->d3dDeviceInterop);
			gl_d3d_buffers->d3dDeviceInterop = NULL;
			add_log("Disabled DX-OGL Interop layer");
		}

		// Release Direct3D objects
		if (gl_d3d_buffers->d3dLeftColorTexture)
		{
			IDirect3DSurface9 *d3dLeftColorTexture = (IDirect3DSurface9 *)gl_d3d_buffers->d3dLeftColorTexture;
			d3dLeftColorTexture->Release();
			gl_d3d_buffers->d3dLeftColorTexture = NULL;
			add_log("Releasing Direct3D Left Color Texture");
		}

		if (gl_d3d_buffers->d3dRightColorTexture)
		{
			IDirect3DSurface9 *d3dRightColorTexture = (IDirect3DSurface9 *)gl_d3d_buffers->d3dRightColorTexture;
			d3dRightColorTexture->Release();
			gl_d3d_buffers->d3dRightColorTexture = NULL;
			add_log("Releasing Direct3D Right Color Texture");
		}

		if (gl_d3d_buffers->d3dLeftColorBuffer)
		{
			IDirect3DSurface9 *d3dLeftColorBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dLeftColorBuffer;
			d3dLeftColorBuffer->Release();
			gl_d3d_buffers->d3dLeftColorBuffer = NULL;
			add_log("Releasing Direct3D Left Color Buffer");
		}

		if (gl_d3d_buffers->d3dRightColorBuffer)
		{
			IDirect3DSurface9 *d3dRightColorBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dRightColorBuffer;
			d3dRightColorBuffer->Release();
			gl_d3d_buffers->d3dRightColorBuffer = NULL;
			add_log("Releasing Direct3D Right Color Buffer");
		}

		if (gl_d3d_buffers->d3dDepthBuffer)
		{
			IDirect3DSurface9 *d3dDepthBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dDepthBuffer;
			d3dDepthBuffer->Release();
			gl_d3d_buffers->d3dDepthBuffer = NULL;
			add_log("Releasing Direct3D Depth Buffer");
		}

		if (gl_d3d_buffers->d3dBackBuffer)
		{
			IDirect3DSurface9 *d3dBackBuffer = (IDirect3DSurface9 *)gl_d3d_buffers->d3dBackBuffer;
			d3dBackBuffer->Release();
			gl_d3d_buffers->d3dBackBuffer = NULL;
			add_log("Releasing Direct3D Back Buffer");
		}

		if (gl_d3d_buffers->d3dDevice)
		{
			IDirect3DDevice9 *d3dDevice = (IDirect3DDevice9 *)gl_d3d_buffers->d3dDevice;
			d3dDevice->Release();
			gl_d3d_buffers->d3dDevice = NULL;
			add_log("Releasing Direct3D Device");
		}

		if (gl_d3d_buffers->d3dLibrary)
		{
			FreeLibrary((HINSTANCE)gl_d3d_buffers->d3dLibrary);
			gl_d3d_buffers->d3dLibrary = NULL;
			add_log("DirectX Library unloaded");
		}

		// Delete OpenGL frame buffer object
		if (gl_d3d_buffers->fbo_left)
		{
			glDeleteFramebuffers(1, &gl_d3d_buffers->fbo_left);
			gl_d3d_buffers->fbo_left = 0;
			add_log("Deleting OpenGL Left FBO");
		}

		// Delete OpenGL render buffer
		if (gl_d3d_buffers->render_buffer)
		{
			glDeleteRenderbuffersEXT(1, &gl_d3d_buffers->render_buffer);
			gl_d3d_buffers->render_buffer = 0;
			add_log("Deleting OpenGL Render Buffer");
		}

		// Delete OpenGL textures
		if (gl_d3d_buffers->texture_left)
		{
			glDeleteTextures(1, &gl_d3d_buffers->texture_left);
			gl_d3d_buffers->texture_left = 0;
			add_log("Deleting OpenGL Left Texture");
		}

		if (gl_d3d_buffers->texture_right)
		{
			glDeleteTextures(1, &gl_d3d_buffers->texture_right);
			gl_d3d_buffers->texture_right = 0;
			add_log("Deleting OpenGL Right Texture");
		}
		// Unload NvAPI
		NvAPI_Unload();
		add_log("Cleanup complete!");
		m_reInit = false;
		m_NV3DVisionIsInInitPhase = true;
	}
}
///--------------------------------------------------------------------------------------

// our NVAPI poll
static void _NVAPIquery(void)
{
	// Run as long as is alive
	while (m_areThreadRunning)
	{
		// Accept keys only if we are initialized
		if (NV3DVisionIsNotInit() == false)
		{
			m_nvapiMutex.lock();
	
			// Key Handler
			/////////////////////////////////////
			size_t currentKey;
			for (currentKey = 0; currentKey < g_reader->GetNumberOfKeys(); currentKey++)
			{
				/////////////////////////////////////
				// Toggle settings
				/////////////////////////////////////
				if (g_reader->GetKeyToggleMode(currentKey))
				{
					// Toggled Enabled
					if (IsAltKeyToggleKeyDown(currentKey) || IsXControllerAltKeyToggleKeyDown(currentKey))
					{
						// Convergence
						if (g_reader->GetAltConvergence(currentKey) != -1.0f)
						{
							m_isCustomConvergenceApplied = !m_isCustomConvergenceApplied;
							// Custom Convergence
							if (m_isCustomConvergenceApplied)
							{
								// Apply the new convergence
								g_convergence = g_reader->GetAltConvergence(currentKey);
								NvAPI_Stereo_SetConvergence(nvStereo, g_convergence);
							}
							else
							{
								NvAPI_Stereo_SetConvergence(nvStereo, m_currentConvergence);
							}
						}
							
						// Separation
						if (g_reader->GetAltSeparation(currentKey) != -1.0f)
						{
							m_isCustomSeparationApplied = !m_isCustomSeparationApplied;
							// Custom Separation
							if (m_isCustomSeparationApplied)
							{
								// Apply the new separation	
								g_separation = g_reader->GetAltSeparation(currentKey);
								g_eyeSeparation = g_separation * g_eyeSeparation / 100;
								NvAPI_Stereo_SetSeparation(nvStereo, g_separation);
							}
							else
							{
								NvAPI_Stereo_SetSeparation(nvStereo, m_currentSeparation);
							}
						}

						// Custom (x,y,z,w);
						CUSTOM_SHADER_VALUES_T customAltShaderValues = { 0, -1.0f, -1.0f, -1.0f, -1.0f };
						if (g_reader->GetAltCustomShaderParams(currentKey, customAltShaderValues) == TRUE)
						{
							// What is the current state of the key??
							bool currentState = g_reader->GetAltCustomShaderApplied(currentKey);

							// Do we have valid values?
							// X
							if (customAltShaderValues.x != -1.0f)
							{
								// Store our read values
								if (currentState == false)
									g_customAltShaderValues.x = customAltShaderValues.x;
								else
									g_customAltShaderValues.x = -1.0f;
								g_reader->SetAltCustomShaderApplied(currentKey, !currentState);
							}
							// Y
							if (customAltShaderValues.y != -1.0f)
							{
								// Store our read values
								if (currentState == false)
									g_customAltShaderValues.y = customAltShaderValues.y;
								else
									g_customAltShaderValues.y = -1.0f;
								g_reader->SetAltCustomShaderApplied(currentKey, !currentState);
							}
							// Z
							if (customAltShaderValues.z != -1.0f)
							{
								// Store our read values
								if (currentState == false)
									g_customAltShaderValues.z = customAltShaderValues.z;
								else
									g_customAltShaderValues.z = -1.0f;
								g_reader->SetAltCustomShaderApplied(currentKey, !currentState);
							}
							// W
							if (customAltShaderValues.w != -1.0f)
							{
								// Store our read values
								if (currentState == false)
									g_customAltShaderValues.w = customAltShaderValues.w;
								else
									g_customAltShaderValues.w = -1.0f;
								g_reader->SetAltCustomShaderApplied(currentKey, !currentState);
							}

							// Do we have any shader being applied???
							// if everything is disabled, clean up the variables
							if (g_reader->GetAltCustomShaderIsAnythingApplied() == false)
							{
								g_customAltShaderValues = { 0, -1.0f, -1.0f, -1.0f, -1.0f };
							}
						}
						break;
					}
				}
				else 
				/////////////////////////////////////
				// Push to hold settings
				// Only Separation & Convergence are handled
				/////////////////////////////////////
				if (IsKeyDown(g_reader->GetKeyNumber(currentKey)))
				{
					// Custom Convergence
					if (g_reader->GetAltConvergence(currentKey) != -1.0f)
					{
						// Apply the new convergence
						g_convergence = g_reader->GetAltConvergence(currentKey);
						NvAPI_Stereo_SetConvergence(nvStereo, g_convergence);
						m_isCustomConvergenceApplied = true;
					}
					// Custom Separation
					if (g_reader->GetAltSeparation(currentKey) != -1.0f)
					{
						// Apply the new separation	
						g_separation = g_reader->GetAltSeparation(currentKey);
						g_eyeSeparation = g_separation * g_eyeSeparation / 100;
						NvAPI_Stereo_SetSeparation(nvStereo, g_separation);
						m_isCustomSeparationApplied = true;
					}
					m_isPressAndHold = true;
					break;
				}
				else if (m_isPressAndHold && g_reader->GetKeyNumber(currentKey))
				{
					NvAPI_Stereo_SetSeparation(nvStereo, m_currentSeparation);
					NvAPI_Stereo_SetConvergence(nvStereo, m_currentConvergence);
					g_reader->SetKeyPrevState(g_reader->GetKeyNumber(currentKey), 0);
					m_isCustomSeparationApplied = false;
					m_isCustomConvergenceApplied = false;
					m_isPressAndHold = false;
				}
			}

			if (m_isCustomSeparationApplied == false)
			{
				// get the correct separation
				NvAPI_Stereo_GetEyeSeparation(nvStereo, &g_eyeSeparation);
				NvAPI_Stereo_GetSeparation(nvStereo, &g_separation);
				g_eyeSeparation = g_separation * g_eyeSeparation / 100;
				float factor = g_reader->GetDepthFactor();
				if (factor > 0.0)
				{
					g_eyeSeparation = g_eyeSeparation * factor;
				}
				// Update the normal convergence & separation values
				m_currentSeparation = g_separation;
			}

			if (m_isCustomConvergenceApplied == false)
			{
				NvAPI_Stereo_GetConvergence(nvStereo, &g_convergence);
				m_currentConvergence = g_convergence;
			}
			
			if ((IsKeyDown(0x11) != 0) && (IsKeyToggleKeyDown(m_debugKeyCodes[1], m_keyStates[1], m_prevKeyStates[1]))) // CTRL + [F10]key Re-Init the renderer when NOT in DEV Mode
			{
				if (!g_reader->GetDevelopmentMode())
				{
					g_windowResizeCheck = true;
				}
			}
			else if ((IsKeyDown(0x11) != 0) && (IsKeyDown(0x12) != 0) && (IsKeyToggleKeyDown(m_debugKeyCodes[2], m_keyStates[2], m_prevKeyStates[2]))) // CTRL + ALT + F11
			{
				// Flip the mode only if Depth3D is enabled
				if (g_reader->Get3DCompatModeEnabled())
					g_reader->Set3DCompatModeActive(!g_reader->Get3DCompatModeActive());
			}
			m_nvapiMutex.unlock();
		}
		Sleep(20);
	}
	add_log("General Monitor Thread Stopped");
}
///--------------------------------------------------------------------------------------

static void _DEV_KeyboardKeys(void)
{
	// Run as long as is alive
	while (m_areThreadRunning)
	{
		// Accept keys only if we are initialized
		if (NV3DVisionIsNotInit() == false)
		{
			if ((IsKeyDown(0x11) != 0) && (IsKeyDown(0xBB) != 0))  // [CTRL +] key next Vertex shader
			{
				if (!KillShaderMessageBox())
				{
					// find first program index in shader list.
					ShaderManager *shaderManager = ShaderManager::getInstance();
					std::vector<EXISTING_SHADER_T> existingShaders = shaderManager->GetExistingShaders();
					size_t index = indexVertex;

					for (; index < (existingShaders.size() - 1); index++)
					{
						if ((debugVertexIndex == (int)existingShaders[index].m_programId))
						{
							// We found the current one
							break;
						}
					}

					// Now we need to find the next valid one
					while ((index++) < (existingShaders.size() - 1))
					{
						int newIndex = existingShaders[index].m_programId;
						if (debugVertexIndex != newIndex)
						{
							debugVertexIndex = (int)existingShaders[index].m_programId;
							indexVertex = index;
							break;
						}
					}

					isCurrentShaderVertex = true;
					printCurrentShader = true;
				}
			}
			else if ((IsKeyDown(0x11) != 0) && IsKeyDown(0xBD) != 0) //[CTRL -] key previous Vertex shader
			{
				if (!KillShaderMessageBox())
				{
					// Now we need to find the next valid one
					ShaderManager *shaderManager = ShaderManager::getInstance();
					std::vector<EXISTING_SHADER_T> existingShaders = shaderManager->GetExistingShaders();
					size_t index = indexVertex;

					for (; index > 1; index--)
					{
						if ((debugVertexIndex == (int)existingShaders[index].m_programId))
						{
							// We found the current one
							break;
						}
					}

					// Now we need to find the next valid one
					while ((index--) > 1)
					{
						int newIndex = existingShaders[index].m_programId;
						if (debugVertexIndex != newIndex)
						{
							debugVertexIndex = (int)existingShaders[index].m_programId;
							indexVertex = index;
							break;
						}
					}
					isCurrentShaderVertex = true;
					printCurrentShader = true;
				}
			}
			else if ((IsKeyDown(0x11) != 0) && (IsKeyDown(0xDD) != 0))  // [CTRL ]] key next Pixel shader
			{
				if (!KillShaderMessageBox())
				{
					// find first program index in shader list.
					ShaderManager *shaderManager = ShaderManager::getInstance();
					std::vector<EXISTING_SHADER_T> existingShaders = shaderManager->GetExistingShaders();
					size_t index = indexPixel;

					for (; index < (existingShaders.size() - 1); index++)
					{
						if ((debugPixelIndex == (int)existingShaders[index].m_programId))
						{
							// We found the current one
							break;
						}
					}

					// Now we need to find the next valid one
					while ((index++) < (existingShaders.size() - 1))
					{
						int newIndex = existingShaders[index].m_programId;
						if (debugPixelIndex != newIndex)
						{
							debugPixelIndex = (int)existingShaders[index].m_programId;
							indexPixel = index;
							break;
						}
					}

					isCurrentShaderVertex = false;
					printCurrentShader = true;
				}
			}
			else if ((IsKeyDown(0x11) != 0) && IsKeyDown(0xDB) != 0) //[CTRL [] key previous Pixel shader
			{
				if (!KillShaderMessageBox())
				{
					// find first program index in shader list.
					ShaderManager *shaderManager = ShaderManager::getInstance();
					std::vector<EXISTING_SHADER_T> existingShaders = shaderManager->GetExistingShaders();
					size_t index = indexPixel;

					for (; index > 1; index--)
					{
						if ((debugPixelIndex == (int)existingShaders[index].m_programId))
						{
							// We found the current one
							break;
						}
					}

					// Now we need to find the next one
					while ((index--) > 1)
					{
						int newIndex = existingShaders[index].m_programId;
						if (debugPixelIndex != newIndex)
						{
							debugPixelIndex = (int)existingShaders[index].m_programId;
							indexPixel = index;
							break;
						}
					}

					isCurrentShaderVertex = false;
					printCurrentShader = true;
				}
			}
			else if (IsKeyToggleKeyDown(m_debugKeyCodes[0], m_keyStates[0], m_prevKeyStates[0])) // [F9]key compile current selected shader
			{
				isCompileEnable = true;
			}
			else if (IsKeyToggleKeyDown(m_debugKeyCodes[1], m_keyStates[1], m_prevKeyStates[1])) // [F10]key Show the current Convergence + Separation Values
			{
				std::thread Print3DVisionInfo(Print3DVisionInfo);
				Print3DVisionInfo.detach();
			}
			else if (IsKeyToggleKeyDown(m_debugKeyCodes[2], m_keyStates[2], m_prevKeyStates[2])) // [F11]key Reload Config file (Shader Wise)
			{
				ShaderManager * g_shaderManager = ShaderManager::getInstance();
				g_shaderManager->ReadConfigFile();
				MessageBox(NULL, "Config File Read !", "Info", MB_OK);
			}
		}
		Sleep(300);
	}
	add_log("DevKeyboard Monitor Thread Stopped");
}
///--------------------------------------------------------------------------------------

static void _NVAPi_3DIsActive(void)
{
	// Run as long as is alive
	while (m_areThreadRunning)
	{
		// Accept keys only if we are initialized
		if (NV3DVisionIsNotInit() == false)
		{
			m_nvapiMutex.lock();

			// Get the state of 3D Vision activation status
			NvAPI_Stereo_IsActivated(nvStereo, &g_is3DVisionActive);

			m_nvapiMutex.unlock();
		}
		// We really don't need to query this very fast AS IS A SLOW FUNCTION!!!
		Sleep(500);
	}
	add_log("NVAPI_R367 Query Thread Stopped");
}
///--------------------------------------------------------------------------------------

static bool KillShaderMessageBox(void)
{
	// Get the info box and kill it
	HWND hwnd = FindWindow(NULL, "Shader Info :");
	if (hwnd)
	{
		SendMessage(hwnd, WM_CLOSE, 0, 0);
		return true;
	}
	return false;
}
///--------------------------------------------------------------------------------------

static void Print3DVisionInfo(void)
{
	char str[500];
	sprintf_s(str, "Current 3D Vision Settings: \n\nSEPARATION (Depth): %f.\n\nCONVERGENCE: %f. \n", Get3DEyeSeparation(), Get3DConvergence());
	
	// Print info
	MessageBox(NULL, str, "3D Vision Current Settings :", MB_OK);
}
///--------------------------------------------------------------------------------------

static bool IsKeyToggleKeyDown(unsigned int keyCode, int &keyState, int &prevState)
{
	//Return if the high byte is true (ie key is down)
	bool ret = false;
	int state = GetAsyncKeyState(keyCode);

	if ((state & 1 && (GetTickCount() - prevState > 100)) && (state != keyState))
	{
		ret = true;
		prevState = GetTickCount();
		keyState = state;
	}
	else
	{
		keyState = state;
	}
	return ret;
}
///--------------------------------------------------------------------------------------

static bool IsAltKeyToggleKeyDown(int keyCodeIndex)
{
	//Return if the high byte is true (ie key is down)
	bool ret = false;
	int state = GetAsyncKeyState(g_reader->GetKeyNumber(keyCodeIndex));
	int keyState = g_reader->GetKeyState(keyCodeIndex);
	int prevState = g_reader->GetKeyPrevState(keyCodeIndex);
	
	if ((state & 1 && (GetTickCount() - prevState > 100)) && (state != keyState))
	{
		ret = true;
		prevState = GetTickCount();
		g_reader->SetKeyPrevState(keyCodeIndex, prevState);
		g_reader->SetKeyState(keyCodeIndex, state);
	}
	else
	{
		g_reader->SetKeyState(keyCodeIndex, state);
	}
	return ret;
}
///--------------------------------------------------------------------------------------

static bool IsXControllerAltKeyToggleKeyDown(int keyCodeIndex)
{
	bool ret = false;
	int currentKey = g_reader->GetKeyNumber(keyCodeIndex);
	if (
		(currentKey == XINPUT_GAMEPAD_DPAD_UP) ||
		(currentKey == XINPUT_GAMEPAD_DPAD_DOWN) ||
		(currentKey == XINPUT_GAMEPAD_DPAD_LEFT) ||
		(currentKey == XINPUT_GAMEPAD_DPAD_RIGHT) ||
		(currentKey == XINPUT_GAMEPAD_START) ||
		(currentKey == XINPUT_GAMEPAD_BACK) ||
		(currentKey == XINPUT_GAMEPAD_LEFT_THUMB) ||
		(currentKey == XINPUT_GAMEPAD_RIGHT_THUMB) ||
		(currentKey == XINPUT_GAMEPAD_LEFT_SHOULDER) ||
		(currentKey == XINPUT_GAMEPAD_RIGHT_SHOULDER) ||
		(currentKey == XINPUT_GAMEPAD_A) ||
		(currentKey == XINPUT_GAMEPAD_B) ||
		(currentKey == XINPUT_GAMEPAD_X) ||
		(currentKey == XINPUT_GAMEPAD_Y)
		)
	{
		//Return if the high byte is true (ie key is down)
		int state = g_xController->GetState().Gamepad.wButtons & g_reader->GetKeyNumber(keyCodeIndex);
		int prevState = g_reader->GetKeyPrevState(keyCodeIndex);

		if ((state) && (state != prevState))
			ret = true;
		prevState = state;
		g_reader->SetKeyPrevState(keyCodeIndex, prevState);
	}
	return ret;
}
///--------------------------------------------------------------------------------------

static bool IsKeyDown(int keyCode)
{
	//Return if the high byte is true (ie key is down)
	return (GetKeyState(keyCode) & 0x80) != 0;
}
///--------------------------------------------------------------------------------------

void Set3DSeparation(float value)
{
	m_nvapiMutex.lock();
	g_separation = value;
	m_nvapiMutex.unlock();
}
///--------------------------------------------------------------------------------------

void Set3DConvergence(float value)
{
	m_nvapiMutex.lock();
	g_convergence = value;
	m_nvapiMutex.unlock();
}
///--------------------------------------------------------------------------------------

void Set3DEyeSeparation(float value)
{
	m_nvapiMutex.lock();
	g_eyeSeparation = value;
	m_nvapiMutex.unlock();
}
///--------------------------------------------------------------------------------------

BOOL Get3DVisionIsActive()
{
	BOOL ret = 0;

	m_nvapiMutex.lock();
	ret = static_cast<BOOL>(g_is3DVisionActive);
	m_nvapiMutex.unlock();

	return ret;
}
///--------------------------------------------------------------------------------------

float Get3DSeparation()
{
	float ret = 0;

	m_nvapiMutex.lock();
	ret = g_separation;
	m_nvapiMutex.unlock();

	return ret;
}
///--------------------------------------------------------------------------------------

float Get3DConvergence()
{
	float ret = 0;

	m_nvapiMutex.lock();
	ret = g_convergence;
	m_nvapiMutex.unlock();

	return ret;
}
///--------------------------------------------------------------------------------------

float Get3DEyeSeparation()
{
	float ret = 0;

	m_nvapiMutex.lock();
	ret = g_eyeSeparation;
	m_nvapiMutex.unlock();

	return ret;
}
///--------------------------------------------------------------------------------------

bool NV3DVisionIsNotInit()
{
	return m_NV3DVisionIsInInitPhase;
}
///--------------------------------------------------------------------------------------

void NV3DVisionSetCurrentFrame(unsigned int frameNumber)
{
	m_frameCall = frameNumber;
}
///--------------------------------------------------------------------------------------

unsigned int NV3DVisionGetCurrentFrame(void)
{
	return m_frameCall;
}
///--------------------------------------------------------------------------------------

void OGL3DVisionWrapperGetVersion(char *pVersionStr)
{
	if (pVersionStr)
	{
		strcpy_s(pVersionStr, 255, OGL3DVISION_VERSION_STR);
	}
}
///--------------------------------------------------------------------------------------

void StartNVAPIThread(void)
{
	if (!m_isResizeCheckThreadInitialised)
	{
		add_log("Starting General Monitor Thread");
		// Start our nvapi thread poll
		std::thread NVAPIPoll(_NVAPIquery);
		NVAPIPoll.detach();

		add_log("Starting NVAPI_R367 Query Thread");
		std::thread NVAPI_3DisActive(_NVAPi_3DIsActive);
		NVAPI_3DisActive.detach();

		if (g_reader->GetDevelopmentMode())
		{
			add_log("Starting DevKeyboard Query Thread");
			std::thread DevKeyboardThread(_DEV_KeyboardKeys);
			DevKeyboardThread.detach();
		}

		m_isResizeCheckThreadInitialised = true;
	}
}
///--------------------------------------------------------------------------------------

void StoptNVAPIThread(void)
{
	if (m_isResizeCheckThreadInitialised)
	{
		m_areThreadRunning = false;
	}
}
///--------------------------------------------------------------------------------------

void NV3DVisionReInitBuffers(void)
{
	add_log("Re-initialising 3D Vision Buffers");
	m_reInit = true;
}
///-------------------------------------------------------------------------------------


